"""
.. _modifying_data:

Adding/Removing Containers from an NWB File
============================================

This tutorial explains how to add and remove containers from an existing NWB file and either write the data back to the
same file or export the data to a new file.
"""

###############################################################################
# Adding objects to an NWB file in read/write mode
# ----------------------------------------------------
# PyNWB supports adding container objects to an existing NWB file - that is, reading data from an NWB file, adding a
# container object, such as a new :py:class:`~pynwb.base.TimeSeries` object, and writing the modified
# :py:class:`~pynwb.file.NWBFile` back to the same file path on disk. To do so:
#
# 1. open the file with an :py:class:`~pynwb.NWBHDF5IO` object in read/write mode (``mode='r+'`` or ``mode='a'``)
# 2. read the :py:class:`~pynwb.file.NWBFile`
# 3. add container objects to the :py:class:`~pynwb.file.NWBFile` object
# 4. write the modified :py:class:`~pynwb.file.NWBFile` using the same :py:class:`~pynwb.NWBHDF5IO` object
#
# For example:

import datetime

import numpy as np

# sphinx_gallery_thumbnail_path = 'figures/gallery_thumbnails_add_remove_containers.png'
from pynwb import NWBHDF5IO, NWBFile, TimeSeries

# first, write a test NWB file
nwbfile = NWBFile(
    session_description="demonstrate adding to an NWB file",
    identifier="NWB123",
    session_start_time=datetime.datetime.now(datetime.timezone.utc),
)

filename = "nwbfile.nwb"
with NWBHDF5IO(filename, "w") as io:
    io.write(nwbfile)

# open the NWB file in r+ mode
with NWBHDF5IO(filename, "r+") as io:
    read_nwbfile = io.read()

    # create a TimeSeries and add it to the file under the acquisition group
    data = list(range(100, 200, 10))
    timestamps = np.arange(10, dtype=float)
    test_ts = TimeSeries(
        name="test_timeseries", data=data, unit="m", timestamps=timestamps
    )
    read_nwbfile.add_acquisition(test_ts)

    # write the modified NWB file
    io.write(read_nwbfile)

# confirm the file contains the new TimeSeries in acquisition
with NWBHDF5IO(filename, "r") as io:
    read_nwbfile = io.read()
    print(read_nwbfile)

###############################################################################
# .. note::
#
#   You cannot remove objects from an NWB file using the above method.

###############################################################################
# Modifying an NWB file in this way has limitations. The destination file path must be the same as the source
# file path, and it is not possible to remove objects from an NWB file. You can use the
# :py:meth:`NWBHDF5IO.export <pynwb.NWBHDF5IO.export>` method, detailed below, to modify an NWB file in these ways.
#
# Exporting a written NWB file to a new file path
# -----------------------------------------------
# Use the :py:meth:`NWBHDF5IO.export <pynwb.NWBHDF5IO.export>` method to read data from an existing NWB file,
# modify the data, and write the modified data to a new file path. Modifications to the data can be additions or
# removals of objects, such as :py:class:`~pynwb.base.TimeSeries` objects. This is especially useful if you
# have raw data and processed data in the same NWB file and you want to create a new NWB file with all the contents of
# the original file except for the raw data for sharing with collaborators.
#
# To remove existing containers, use the :py:meth:`~hdmf.utils.LabelledDict.pop` method on any
# :py:class:`~hdmf.utils.LabelledDict` object, such as ``NWBFile.acquisition``, ``NWBFile.processing``,
# ``NWBFile.analysis``, ``NWBFile.processing``, ``NWBFile.scratch``, ``NWBFile.devices``, ``NWBFile.stimulus``,
# ``NWBFile.stimulus_template``, ``NWBFile.electrode_groups``, ``NWBFile.imaging_planes``,
# ``NWBFile.icephys_electrodes``, ``NWBFile.ogen_sites``, ``NWBFile.lab_meta_data``,
# and :py:class:`~pynwb.base.ProcessingModule` objects.
#
# For example:

# first, create a test NWB file with a TimeSeries in the acquisition group
nwbfile = NWBFile(
    session_description="demonstrate export of an NWB file",
    identifier="NWB123",
    session_start_time=datetime.datetime.now(datetime.timezone.utc),
)
data1 = list(range(100, 200, 10))
timestamps1 = np.arange(10, dtype=float)
test_ts1 = TimeSeries(
    name="test_timeseries1", data=data1, unit="m", timestamps=timestamps1
)
nwbfile.add_acquisition(test_ts1)

# then, create a processing module for processed behavioral data
nwbfile.create_processing_module(
    name="behavior", description="processed behavioral data"
)
data2 = list(range(100, 200, 10))
timestamps2 = np.arange(10, dtype=float)
test_ts2 = TimeSeries(
    name="test_timeseries2", data=data2, unit="m", timestamps=timestamps2
)
nwbfile.processing["behavior"].add(test_ts2)

# write these objects to an NWB file
filename = "nwbfile.nwb"
with NWBHDF5IO(filename, "w") as io:
    io.write(nwbfile)

# read the written file
export_filename = "exported_nwbfile.nwb"
with NWBHDF5IO(filename, mode="r") as read_io:
    read_nwbfile = read_io.read()

    # add a new TimeSeries to the behavior processing module
    data3 = list(range(100, 200, 10))
    timestamps3 = np.arange(10, dtype=float)
    test_ts3 = TimeSeries(
        name="test_timeseries3", data=data3, unit="m", timestamps=timestamps3
    )
    read_nwbfile.processing["behavior"].add(test_ts3)

    # use the pop method to remove the original TimeSeries from the acquisition group
    read_nwbfile.acquisition.pop("test_timeseries1")

    # use the pop method to remove a TimeSeries from a processing module
    read_nwbfile.processing["behavior"].data_interfaces.pop("test_timeseries2")

    # call the export method to write the modified NWBFile instance to a new file path.
    # the original file is not modified
    with NWBHDF5IO(export_filename, mode="w") as export_io:
        export_io.export(src_io=read_io, nwbfile=read_nwbfile)

# confirm the exported file does not contain TimeSeries with names 'test_timeseries1' or 'test_timeseries2'
# but does contain a new TimeSeries in processing['behavior'] with name 'test_timeseries3'
with NWBHDF5IO(export_filename, "r") as io:
    read_nwbfile = io.read()
    print(read_nwbfile)
    print(read_nwbfile.processing["behavior"])

###############################################################################
# .. note::
#
#   :py:class:`~pynwb.epoch.TimeIntervals` objects, such as ``NWBFile.epochs``, ``NWBFile.trials``,
#   ``NWBFile.invalid_times``, and custom :py:class:`~pynwb.epoch.TimeIntervals` objects cannot be
#   removed (popped) from ``NWBFile.intervals``.

###############################################################################
# .. warning::
#
#   Removing an object from an NWBFile may break links and references within the file and across files.
#   This is analogous to having shortcuts/aliases to a file on your filesystem and then deleting the file.
#   Extra caution should be taken when removing heavily referenced items such as
#   :py:class:`~pynwb.device.Device` objects,
#   :py:class:`~pynwb.ecephys.ElectrodeGroup` objects, the electrodes table, and the
#   :py:class:`~pynwb.ophys.PlaneSegmentation` table.

###############################################################################
# Exporting with new object IDs
# ---------------------------------
# When exporting a read NWB file to a new file path, the object IDs within the original NWB file will be copied to the
# new file. To make the exported NWB file contain a new set of object IDs, call
# :py:meth:`~hdmf.container.AbstractContainer.generate_new_id` on your :py:class:`~pynwb.file.NWBFile` object.
# This will generate a new object ID for the :py:class:`~pynwb.file.NWBFile` object and all of the objects within
# the NWB file.

export_filename = "exported_nwbfile.nwb"
with NWBHDF5IO(filename, mode="r") as read_io:
    read_nwbfile = read_io.read()
    read_nwbfile.generate_new_id()

    with NWBHDF5IO(export_filename, mode="w") as export_io:
        export_io.export(src_io=read_io, nwbfile=read_nwbfile)

###############################################################################
# For more information about the export functionality, see :ref:`export`
# and the PyNWB documentation for :py:meth:`NWBHDF5IO.export <pynwb.NWBHDF5IO.export>`.
#
# For more information about editing a file in place, see :ref:`editing`.



from hdmf.build import LinkBuilder
from hdmf.common.table import VectorData
from hdmf.common.io.table import DynamicTableMap

from .. import register_map
from pynwb.epoch import TimeIntervals
from pynwb.base import TimeSeriesReferenceVectorData


@register_map(TimeIntervals)
class TimeIntervalsMap(DynamicTableMap):

    @DynamicTableMap.constructor_arg('columns')
    def columns_carg(self, builder, manager):
        # handle the case when a TimeIntervals is read with a non-TimeSeriesReferenceVectorData "timeseries" column
        # this is the case for NWB schema v2.4 and earlier, where the timeseries column was a regular VectorData.
        timeseries_builder = builder.get('timeseries')

        # handle the case when the TimeIntervals has a "timeseries" column that is a link (e.g., it exists in
        # a different file that was shallow copied to this file).
        if isinstance(timeseries_builder, LinkBuilder):
            timeseries_builder = timeseries_builder.builder

        # if we have a timeseries column and the type is VectorData instead of TimeSeriesReferenceVectorData
        if (timeseries_builder is not None and
                timeseries_builder.attributes['neurodata_type'] != 'TimeSeriesReferenceVectorData'):
            # override builder attributes
            timeseries_builder.attributes['neurodata_type'] = 'TimeSeriesReferenceVectorData'
            timeseries_builder.attributes['namespace'] = 'core'
            # construct new columns list
            columns = list()
            for dset_builder in builder.datasets.values():
                dset_obj = manager.construct(dset_builder)  # these have already been constructed
                # go through only the column datasets and replace the 'timeseries' column class in-place.
                if isinstance(dset_obj, VectorData):
                    if dset_obj.name == 'timeseries':
                        # replacing the class in-place is possible because the VectorData and
                        # TimeSeriesReferenceVectorData have the same memory layout and the old and new
                        # schema are compatible (i.e., only the neurodata_type was changed in 2.5)
                        dset_obj.__class__ = TimeSeriesReferenceVectorData
                        # Execute init logic specific for TimeSeriesReferenceVectorData
                        dset_obj._init_internal()
                    columns.append(dset_obj)
            # overwrite the columns constructor argument
            return columns
        # do not override
        return None



import os
from pynwb.base import TimeSeriesReference
from pynwb import NWBHDF5IO
from pynwb.testing import TestCase
from pynwb.testing.mock.file import mock_NWBFile
from pynwb.testing.mock.base import mock_TimeSeries


class TestFileCopy(TestCase):

    def setUp(self):
        self.path1 = "test_a.h5"
        self.path2 = "test_b.h5"

    def tearDown(self):
        if os.path.exists(self.path1):
            os.remove(self.path1)
        if os.path.exists(self.path2):
            os.remove(self.path2)

    def test_copy_file_link_timeintervals_timeseries(self):
        """Test copying a file with a TimeSeriesReference in a TimeIntervals object and reading that copy.

        Based on https://github.com/NeurodataWithoutBorders/pynwb/issues/1863
        """
        new_nwb = mock_NWBFile()
        test_ts = mock_TimeSeries(name="test_ts", timestamps=[1.0, 2.0, 3.0], data=[1.0, 2.0, 3.0])
        new_nwb.add_acquisition(test_ts)
        new_nwb.add_trial(start_time=1.0, stop_time=2.0, timeseries=[test_ts])

        with NWBHDF5IO(self.path1, 'w') as io:
            io.write(new_nwb)

        with NWBHDF5IO(self.path1, 'r') as base_io:
            # the TimeIntervals object is copied but the TimeSeriesReferenceVectorData is linked
            nwb_add = base_io.read().copy()
            with NWBHDF5IO(self.path2, 'w', manager=base_io.manager) as out_io:
                out_io.write(nwb_add)

        with NWBHDF5IO(self.path2, 'r') as io:
            nwb = io.read()
            ts_val = nwb.trials["timeseries"][0][0]
            assert isinstance(ts_val, TimeSeriesReference)
            assert ts_val.timeseries is nwb.acquisition["test_ts"]



import warnings

import numpy as np

from pynwb import TimeSeries
from pynwb.base import Image, Images, ImageReferences
from pynwb.device import Device
from pynwb.image import (
    ImageSeries,
    IndexSeries,
    ImageMaskSeries,
    OpticalSeries,
    GrayscaleImage,
    RGBImage,
    RGBAImage,
)
from pynwb.testing import TestCase


class ImageSeriesConstructor(TestCase):

    def test_init(self):
        dev = Device('test_device')
        iS = ImageSeries(
            name='test_iS',
            unit='unit',
            external_file=['external_file'],
            starting_frame=[0],
            format='external',
            timestamps=[1., 2., 3.],
            device=dev,
        )
        self.assertEqual(iS.name, 'test_iS')
        self.assertEqual(iS.unit, 'unit')
        self.assertEqual(iS.external_file, ['external_file'])
        self.assertEqual(iS.starting_frame, [0])
        self.assertEqual(iS.format, 'external')
        self.assertIs(iS.device, dev)
        # self.assertEqual(iS.bits_per_pixel, np.nan)

    def test_no_data_no_file(self):
        msg = "Must supply either external_file or data to ImageSeries 'test_iS'."
        with self.assertRaisesWith(ValueError, msg):
            ImageSeries(
                name='test_iS',
                unit='unit',
                rate=2.,
            )

    def test_external_file_no_frame(self):
        iS = ImageSeries(
            name='test_iS',
            unit='unit',
            external_file=['external_file'],
            format="external",
            timestamps=[1., 2., 3.]
        )
        self.assertListEqual(iS.starting_frame, [0])

    def test_data_no_frame(self):
        iS = ImageSeries(
            name='test_iS',
            unit='unit',
            data=np.ones((3, 3, 3)),
            timestamps=[1., 2., 3.]
        )
        self.assertIsNone(iS.starting_frame)

    def test_data_no_unit(self):
        msg = "Must supply 'unit' argument when supplying 'data' to ImageSeries 'test_iS'."
        with self.assertRaisesWith(ValueError, msg):
            ImageSeries(
                name='test_iS',
                data=np.ones((3, 3, 3)),
                timestamps=list()
            )

    def test_external_file_no_unit(self):
        iS = ImageSeries(
            name='test_iS',
            external_file=['external_file'],
            format="external",
            timestamps=list()
        )
        self.assertEqual(iS.unit, ImageSeries.DEFAULT_UNIT)

    def test_dimension_warning(self):
        """Test that a warning is raised when the dimensions of the data are not the
        same as the dimensions of the timestamps."""
        msg = (
            "ImageSeries 'test_iS': Length of data does not match length of timestamps. Your data may be "
            "transposed. Time should be on the 0th dimension"
        )
        with self.assertWarnsWith(UserWarning, msg):
            ImageSeries(
                name='test_iS',
                data=np.ones((3, 3, 3)),
                unit='Frames',
                timestamps=[1, 2, 3, 4]
            )

    def test_dimension_warning_external_file_with_timestamps(self):
        """Test that a warning is not raised when external file is used with timestamps."""
        with warnings.catch_warnings(record=True) as w:
            ImageSeries(
                name='test_iS',
                external_file=['external_file'],
                format='external',
                unit='Frames',
                starting_frame=[0],
                timestamps=[1, 2, 3, 4]
            )
            self.assertEqual(w, [])

    def test_dimension_warning_external_file_with_rate(self):
        """Test that a warning is not raised when external file is used with rate."""
        with warnings.catch_warnings(record=True) as w:
            ImageSeries(
                name='test_iS',
                external_file=['external_file'],
                format='external',
                unit='Frames',
                starting_frame=[0],
                rate=0.2,
            )
            self.assertEqual(w, [])

    def test_external_file_with_incorrect_starting_frame(self):
        """Test that ValueError is raised when the length of starting_frame
        is not the same as the length of external_file."""
        for starting_frame in [None, [0], [1, 2, 3]]:
            with self.subTest():
                msg = (
                    "ImageSeries 'test_iS': The number of frame indices in "
                    "'starting_frame' should have the same length as 'external_file'."
                )
                with self.assertRaisesWith(ValueError, msg):
                    ImageSeries(
                        name="test_iS",
                        external_file=["external_file", "external_file2"],
                        format="external",
                        unit="n.a.",
                        starting_frame=starting_frame,
                        rate=0.2,
                    )

    def test_external_file_with_incorrect_starting_frame_construct_mode(self):
        """Test that warning is raised when the length of starting_frame
        is not the same as the length of external_file in case that the
        ImageSeries in construct mode (i.e., during read)."""
        for starting_frame in [None, [0], [1, 2, 3]]:
            with self.subTest():
                msg = (
                    "ImageSeries 'test_iS': The number of frame indices in "
                    "'starting_frame' should have the same length as 'external_file'."
                )
                # Create the image series in construct mode, modelling the behavior
                # of the ObjectMapper on read while avoiding having to create, write,
                # and read and entire NWB file
                obj = ImageSeries.__new__(ImageSeries,
                                          container_source=None,
                                          parent=None,
                                          object_id="test",
                                          in_construct_mode=True)
                with self.assertWarnsWith(warn_type=UserWarning, exc_msg=msg):
                    obj.__init__(
                        name="test_iS",
                        external_file=["external_file", "external_file2"],
                        format="external",
                        unit="n.a.",
                        starting_frame=starting_frame,
                        rate=0.2,
                    )
            # Disable construct mode. Since we are not using this object any more
            # this is not strictly necessary but is good style in case we expand
            # the test later on
            obj._in_construct_mode = False

    def test_external_file_with_correct_starting_frame(self):
        """Test that ValueError is not raised when the length of starting_frame
        is the same as the length of external_file."""
        with warnings.catch_warnings(record=True) as w:
            ImageSeries(
                name="test_iS",
                external_file=["external_file", "external_file2", "external_file3"],
                format="external",
                unit="n.a.",
                starting_frame=[0, 15, 30],
                rate=0.2,
            )
            self.assertEqual(w, [])

    def test_external_file_with_default_starting_frame(self):
        """Test that starting_frame is set to [0] if not provided, when external_file is has length 1."""
        iS = ImageSeries(
                name="test_iS",
                external_file=["external_file"],
                format="external",
                unit="n.a.",
                starting_frame=None,
                rate=0.2,
        )
        self.assertEqual(iS.starting_frame, [0])

    def test_external_file_with_incorrect_format(self):
        """Test that ValueError is raised when external_file is provided but
        the format is not 'external'."""
        msg = (
            "ImageSeries 'test_iS': Format must be 'external' when external_file is specified."
        )
        with self.assertRaisesWith(ValueError, msg):
            ImageSeries(
                name="test_iS",
                external_file=["external_file"],
                format="raw",
                unit="n.a.",
                starting_frame=[0],
                rate=0.2,
            )

    def test_external_file_with_incorrect_format_construct_mode(self):
        """Test that UserWarning is raised when external_file is provided but
        the format is not 'external' while being in construct mode (i.e,. on data read)."""
        obj = ImageSeries.__new__(ImageSeries,
                                  container_source=None,
                                  parent=None,
                                  object_id="test",
                                  in_construct_mode=True)
        msg = (
            "ImageSeries 'test_iS': Format must be 'external' when external_file is specified."
        )
        with self.assertWarnsWith(warn_type=UserWarning, exc_msg=msg):
            obj.__init__(
                name="test_iS",
                external_file=["external_file"],
                format="raw",
                unit="n.a.",
                starting_frame=[0],
                rate=0.2,
            )

    def test_external_file_default_format(self):
        """Test that format is set to 'external' if not provided, when external_file is provided."""
        msg = (
            "ImageSeries 'test_iS': The value for 'format' has been changed to 'external'. "
            "Setting a default value for 'format' is deprecated and will be changed to "
            "raising a ValueError in the next major release."
        )
        with self.assertWarnsWith(DeprecationWarning, msg):
            iS = ImageSeries(
                name="test_iS",
                external_file=["external_file", "external_file2"],
                unit="n.a.",
                starting_frame=[0, 10],
                rate=0.2,
            )
        self.assertEqual(iS.format, "external")

    def test_external_file_with_correct_format(self):
        """Test that warning is not raised when external_file is provided and format
        is external."""
        with warnings.catch_warnings(record=True) as w:
            ImageSeries(
                name="test_iS",
                external_file=["external_file"],
                format="external",
                unit="n.a.",
                starting_frame=[0],
                rate=0.2,
            )
            self.assertEqual(w, [])

    def test_external_file_with_data(self):
        """Test that ValueError is raised when external_file is provided and
        data is not None."""
        msg = (
            "ImageSeries 'test_iS': Either external_file or data must be specified (not None), but not both."
        )
        with self.assertRaisesWith(ValueError, msg):
            ImageSeries(
                name="test_iS",
                external_file=["external_file"],
                data=np.ones((3, 3, 3)),
                format="external",
                unit="n.a.",
                starting_frame=[0],
                rate=0.2,
            )

    def test_external_file_with_data_construct_mode(self):
        """Test that UserWarning is raised when external_file is provided and
        data is not None while being in construct mode (i.e,. on read)."""
        obj = ImageSeries.__new__(ImageSeries,
                                  container_source=None,
                                  parent=None,
                                  object_id="test",
                                  in_construct_mode=True)
        msg = (
            "ImageSeries 'test_iS': Either external_file or data must be specified (not None), but not both."
        )
        with self.assertWarnsWith(warn_type=UserWarning, exc_msg=msg):
            obj.__init__(
                name="test_iS",
                external_file=["external_file"],
                data=np.ones((3, 3, 3)),
                format="external",
                unit="n.a.",
                starting_frame=[0],
                rate=0.2,
            )


class IndexSeriesConstructor(TestCase):

    def test_init(self):
        image1 = Image(name='test_image', data=np.ones((10, 10)))
        image2 = Image(name='test_image2', data=np.ones((10, 10)))
        image_references = ImageReferences(name='order_of_images', data=[image2, image1])
        images = Images(name='images_name', images=[image1, image2], order_of_images=image_references)

        iS = IndexSeries(
            name='test_iS',
            data=[1, 2, 3],
            unit='N/A',
            indexed_images=images,
            timestamps=[0.1, 0.2, 0.3]
        )
        self.assertEqual(iS.name, 'test_iS')
        self.assertEqual(iS.unit, 'N/A')
        self.assertIs(iS.indexed_images, images)

    def test_init_bad_unit(self):
        ts = TimeSeries(
            name='test_ts',
            data=[1, 2, 3],
            unit='unit',
            timestamps=[0.1, 0.2, 0.3]
        )
        msg = ("The 'unit' field of IndexSeries is fixed to the value 'N/A' starting in NWB 2.5. Passing "
               "a different value for 'unit' will raise an error in PyNWB 3.0.")
        with self.assertWarnsWith(PendingDeprecationWarning, msg):
            iS = IndexSeries(
                name='test_iS',
                data=[1, 2, 3],
                unit='na',
                indexed_timeseries=ts,
                timestamps=[0.1, 0.2, 0.3]
            )
        self.assertEqual(iS.unit, 'N/A')

    def test_init_indexed_ts(self):
        ts = TimeSeries(
            name='test_ts',
            data=[1, 2, 3],
            unit='unit',
            timestamps=[0.1, 0.2, 0.3]
        )
        msg = ("The indexed_timeseries field of IndexSeries is discouraged and will be deprecated in "
               "a future version of NWB. Use the indexed_images field instead.")
        with self.assertWarnsWith(PendingDeprecationWarning, msg):
            iS = IndexSeries(
                name='test_iS',
                data=[1, 2, 3],
                unit='N/A',
                indexed_timeseries=ts,
                timestamps=[0.1, 0.2, 0.3]
            )
        self.assertIs(iS.indexed_timeseries, ts)


class ImageMaskSeriesConstructor(TestCase):

    def test_init(self):
        iS = ImageSeries(name='test_iS', unit='unit',
                         external_file=['external_file'], starting_frame=[0], format='external',
                         timestamps=[1., .2])

        ims = ImageMaskSeries(name='test_ims', unit='unit',
                              masked_imageseries=iS, external_file=['external_file'], starting_frame=[0],
                              format='external', timestamps=[1., 2.])
        self.assertEqual(ims.name, 'test_ims')
        self.assertEqual(ims.unit, 'unit')
        self.assertIs(ims.masked_imageseries, iS)
        self.assertEqual(ims.external_file, ['external_file'])
        self.assertEqual(ims.starting_frame, [0])
        self.assertEqual(ims.format, 'external')


class OpticalSeriesConstructor(TestCase):

    def test_init(self):
        ts = OpticalSeries(name='test_ts', unit='unit', distance=1.0,
                           field_of_view=[4, 5], orientation='orientation', external_file=['external_file'],
                           starting_frame=[0], format='external', timestamps=[1., 2.])
        self.assertEqual(ts.name, 'test_ts')
        self.assertEqual(ts.unit, 'unit')
        self.assertEqual(ts.distance, 1.0)
        self.assertEqual(ts.field_of_view, [4, 5])
        self.assertEqual(ts.orientation, 'orientation')
        self.assertEqual(ts.external_file, ['external_file'])
        self.assertEqual(ts.starting_frame, [0])
        self.assertEqual(ts.format, 'external')


class TestImageSubtypes(TestCase):

    def test_grayscale_image(self):
        GrayscaleImage(name='test_grayscale_image', data=np.ones((2, 2)))

    def test_rgb_image(self):
        RGBImage(name='test_rgb_image', data=np.ones((2, 2, 3)))

    def test_rgba_image(self):
        RGBAImage('test_rgba_image', np.ones((2, 2, 4)))



from pynwb import NWBHDF5IO

from pynwb.testing.mock.file import mock_Subject, mock_NWBFile

from pynwb.testing.mock.base import mock_TimeSeries

from pynwb.testing.mock.ophys import (
    mock_ImagingPlane,
    mock_OnePhotonSeries,
    mock_TwoPhotonSeries,
    mock_RoiResponseSeries,
    mock_PlaneSegmentation,
    mock_OpticalChannel,
    mock_Fluorescence,
    mock_DfOverF,
    mock_ImageSegmentation,
)

from pynwb.testing.mock.ogen import (
    mock_OptogeneticStimulusSite,
    mock_OptogeneticSeries
)

from pynwb.testing.mock.device import mock_Device

from pynwb.testing.mock.behavior import (
    mock_Position,
    mock_PupilTracking,
    mock_CompassDirection,
    mock_SpatialSeries,
)

from pynwb.testing.mock.ecephys import (
    mock_ElectrodeGroup,
    mock_ElectrodeTable,
    mock_ElectricalSeries,
    mock_SpikeEventSeries,
    mock_Units,
)

from pynwb.testing.mock.icephys import (
    mock_IntracellularElectrode,
    mock_CurrentClampSeries,
    mock_IZeroClampSeries,
    mock_VoltageClampSeries,
    mock_VoltageClampStimulusSeries,
    mock_CurrentClampStimulusSeries,
    mock_IntracellularRecordingsTable,
)

import pytest

from pynwb.testing.mock.utils import name_generator, name_generator_registry

mock_functions = [
    mock_ImagingPlane,
    mock_OnePhotonSeries,
    mock_TwoPhotonSeries,
    mock_RoiResponseSeries,
    mock_PlaneSegmentation,
    mock_OpticalChannel,
    mock_Fluorescence,
    mock_DfOverF,
    mock_ImageSegmentation,
    mock_OptogeneticStimulusSite,
    mock_OptogeneticSeries,
    mock_Device,
    mock_Position,
    mock_PupilTracking,
    mock_CompassDirection,
    mock_SpatialSeries,
    mock_ElectrodeGroup,
    mock_ElectrodeTable,
    mock_ElectricalSeries,
    mock_SpikeEventSeries,
    mock_Subject,
    mock_NWBFile,
    mock_TimeSeries,
    mock_CurrentClampSeries,
    mock_IZeroClampSeries,
    mock_VoltageClampSeries,
    mock_VoltageClampStimulusSeries,
    mock_IntracellularElectrode,
    mock_CurrentClampStimulusSeries,
    mock_IntracellularRecordingsTable,
    mock_Units,
]


@pytest.mark.parametrize("mock_function", mock_functions)
def test_mock(mock_function):
    mock_function()


def test_mock_TimeSeries_w_timestamps():
    ts = mock_TimeSeries(timestamps=[0, 1, 2, 3])
    assert ts.timestamps is not None
    assert len(ts.timestamps) == 4


def test_mock_TimeSeries_w_no_time():
    ts = mock_TimeSeries()
    assert ts.rate == 10.0


@pytest.mark.parametrize("mock_function", mock_functions)
def test_mock_write(mock_function, tmp_path):
    if mock_function is mock_NWBFile:
        return
    nwbfile = mock_NWBFile()
    assert mock_function(nwbfile=nwbfile) is not None

    test_file = tmp_path / (mock_function.__name__ + ".nwb")
    with NWBHDF5IO(test_file, "w") as io:
        io.write(nwbfile)


def test_name_generator():

    name_generator_registry.clear()  # reset registry

    assert name_generator("TimeSeries") == "TimeSeries"
    assert name_generator("TimeSeries") == "TimeSeries2"




"""
.. _time_intervals:

Annotating Time Intervals
=========================

Annotating events in time is a common need in neuroscience, e.g. to describes epochs, trials, and
invalid times during an experimental session. NWB supports annotation of time intervals via the
:py:class:`~pynwb.epoch.TimeIntervals` type. The :py:class:`~pynwb.epoch.TimeIntervals` type is
a :py:class:`~hdmf.common.table.DynamicTable` with the following columns:

1. ``start_time`` and ``stop_time`` describe the start and stop times of intervals as floating point offsets in seconds
   relative to the :py:meth:`~pynwb.file.NWBFile.timestamps_reference_time` of the file. In addition,
2. ``tags`` is an optional, indexed column used to associate user-defined string tags with intervals (0 or more tags per
   time interval)
3. ``timeseries`` is an optional, indexed :py:class:`~pynwb.base.TimeSeriesReferenceVectorData` column to map intervals
   directly to ranges in select, relevant :py:class:`~pynwb.base.TimeSeries` (0 or more per time interval)
4. as a :py:class:`~hdmf.common.table.DynamicTable` user may add additional columns to
   :py:meth:`~pynwb.epoch.TimeIntervals` via :py:meth:`~hdmf.common.table.DynamicTable.add_column`


.. hint:: :py:meth:`~pynwb.epoch.TimeIntervals` is intended for storing general annotations of time ranges.
          Depending on the application (e.g., when intervals are generated by data acquisition or automatic
          data processing), it can be useful to describe intervals (or instantaneous events) in time
          as :py:class:`~pynwb.base.TimeSeries`. NWB provides several types for this purposes, e.g.,
          :py:class:`~pynwb.misc.IntervalSeries`, :py:class:`~pynwb.behavior.BehavioralEpochs`,
          :py:class:`~pynwb.behavior.BehavioralEvents`, :py:class:`~pynwb.ecephys.EventDetection`, or
          :py:class:`~pynwb.ecephys.SpikeEventSeries`.

"""

####################
# Setup: Creating an example NWB file for the tutorial
# ----------------------------------------------------
#

# sphinx_gallery_thumbnail_path = 'figures/gallery_thumbnails_timeintervals.png'
from datetime import datetime
from uuid import uuid4

import numpy as np
from dateutil.tz import tzlocal

from pynwb import NWBFile, TimeSeries

# create the NWBFile
nwbfile = NWBFile(
    session_description="my first synthetic recording",  # required
    identifier=str(uuid4()),  # required
    session_start_time=datetime(2017, 4, 3, 11, tzinfo=tzlocal()),  # required
    experimenter="Baggins, Bilbo",  # optional
    lab="Bag End Laboratory",  # optional
    institution="University of Middle Earth at the Shire",  # optional
    experiment_description="I went on an adventure with thirteen dwarves to reclaim vast treasures.",  # optional
    session_id="LONELYMTN",  # optional
)
# create some example TimeSeries
test_ts = TimeSeries(
    name="series1",
    data=np.arange(1000),
    unit="m",
    timestamps=np.linspace(0.5, 601, 1000),
)
rate_ts = TimeSeries(
    name="series2", data=np.arange(600), unit="V", starting_time=0.0, rate=1.0
)
# Add the TimeSeries to the file
nwbfile.add_acquisition(test_ts)
nwbfile.add_acquisition(rate_ts)

####################
# Adding Time Intervals to a NWBFile
# ----------------------------------
#
# NWB provides a set of pre-defined :py:class:`~pynwb.epoch.TimeIntervals`
# tables for :py:meth:`~pynwb.file.NWBFile.epochs`, :py:meth:`~pynwb.file.NWBFile.trials`, and
# :py:meth:`~pynwb.file.NWBFile.invalid_times`.
#
# .. _trials:
#
# Trials
# ^^^^^^
#
# Trials can be added to an NWB file using the methods :py:meth:`~pynwb.file.NWBFile.add_trial`
# By default, NWBFile only requires trial ``start_time`` and ``stop_time``. The ``tags`` and ``timeseries`` are
# optional. For ``timeseries`` we only need to supply the :py:class:`~pynwb.base.TimeSeries`.
# PyNWB automatically calculates the corresponding index range (described by ``idx_start`` and ``count``) for
# the supplied  :py:class:`~pynwb.base.TimeSeries` based on the given ``start_time`` and ``stop_time`` and
# the :py:meth:`~pynwb.base.TimeSeries.timestamps` (or :py:class:`~pynwb.base.TimeSeries.starting_time`
# and :py:meth:`~pynwb.base.TimeSeries.rate`) of the given :py:class:`~pynwb.base.TimeSeries`.
#
# Additional columns can be added using :py:meth:`~pynwb.file.NWBFile.add_trial_column`. This method takes a name
# for the column and a description of what the column stores. You do not need to supply data
# type, as this will inferred. Once all columns have been added, trial data can be populated using
# :py:meth:`~pynwb.file.NWBFile.add_trial`. Note that if you add a custom column, you must
# add at least one row to write the table to a file.
#
# Lets add an additional column and some trial data with tags and timeseries references.

nwbfile.add_trial_column(name="stim", description="the visual stimuli during the trial")

nwbfile.add_trial(
    start_time=0.0,
    stop_time=2.0,
    stim="dog",
    tags=["animal"],
    timeseries=[test_ts, rate_ts],
)
nwbfile.add_trial(
    start_time=3.0,
    stop_time=5.0,
    stim="mountain",
    tags=["landscape"],
    timeseries=[test_ts, rate_ts],
)
nwbfile.add_trial(
    start_time=6.0,
    stop_time=8.0,
    stim="desert",
    tags=["landscape"],
    timeseries=[test_ts, rate_ts],
)
nwbfile.add_trial(
    start_time=9.0,
    stop_time=11.0,
    stim="tree",
    tags=["landscape", "plant"],
    timeseries=[test_ts, rate_ts],
)
nwbfile.add_trial(
    start_time=12.0,
    stop_time=14.0,
    stim="bird",
    tags=["animal"],
    timeseries=[test_ts, rate_ts],
)
nwbfile.add_trial(
    start_time=15.0,
    stop_time=17.0,
    stim="flower",
    tags=["animal"],
    timeseries=[test_ts, rate_ts],
)

####################
# Epochs
# ^^^^^^
#
# Similarly, epochs can be added to an NWB file using the method :py:meth:`~pynwb.file.NWBFile.add_epoch` and
# :py:meth:`~pynwb.file.NWBFile.add_epoch_column`.

nwbfile.add_epoch(
    2.0,
    4.0,
    ["first", "example"],
    [
        test_ts,
    ],
)
nwbfile.add_epoch(
    6.0,
    8.0,
    ["second", "example"],
    [
        test_ts,
    ],
)

####################
# Invalid Times
# ^^^^^^^^^^^^^
#
# Similarly, invalid times can be added using the method :py:meth:`~pynwb.file.NWBFile.add_invalid_time_interval` and
# :py:meth:`~pynwb.file.NWBFile.add_invalid_times_column`.

nwbfile.add_epoch(
    2.0,
    4.0,
    ["first", "example"],
    [
        test_ts,
    ],
)
nwbfile.add_epoch(
    6.0,
    8.0,
    ["second", "example"],
    [
        test_ts,
    ],
)

####################
# Custom Time Intervals
# ^^^^^^^^^^^^^^^^^^^^^
#
# To define custom, experiment-specific :py:class:`~pynwb.epoch.TimeIntervals` we can add them
# either: 1) when creating the :py:class:`~pynwb.file.NWBFile` by defining the
# ``intervals`` constructor argument or 2) via the
# :py:meth:`~pynwb.file.NWBFile.add_time_intervals` or :py:meth:`~pynwb.file.NWBFile.create_time_intervals`
# after the :py:class:`~pynwb.file.NWBFile` has been created.
#

from pynwb.epoch import TimeIntervals

sleep_stages = TimeIntervals(
    name="sleep_stages",
    description="intervals for each sleep stage as determined by EEG",
)

sleep_stages.add_column(name="stage", description="stage of sleep")
sleep_stages.add_column(name="confidence", description="confidence in stage (0-1)")

sleep_stages.add_row(start_time=0.3, stop_time=0.5, stage=1, confidence=0.5)
sleep_stages.add_row(start_time=0.7, stop_time=0.9, stage=2, confidence=0.99)
sleep_stages.add_row(start_time=1.3, stop_time=3.0, stage=3, confidence=0.7)

_ = nwbfile.add_time_intervals(sleep_stages)


####################
# Accessing Time Intervals
# ------------------------
#
# We can access the predefined :py:class:`~pynwb.epoch.TimeIntervals` tables via the corresponding
# :py:meth:`~pynwb.file.NWBFile.epochs`, :py:meth:`~pynwb.file.NWBFile.trials`, and
# :py:meth:`~pynwb.file.NWBFile.invalid_times` properties and for custom :py:class:`~pynwb.epoch.TimeIntervals`
# via the :py:meth:`~pynwb.file.NWBFile.get_time_intervals`  method. E.g.:

_ = nwbfile.intervals
_ = nwbfile.get_time_intervals("sleep_stages")


####################
# Like any :py:class:`~hdmf.common.table.DynamicTable`, we can conveniently convert any
# :py:class:`~pynwb.epoch.TimeIntervals` table to a ``pandas.DataFrame`` via
# :py:meth:`~hdmf.common.table.DynamicTable.to_dataframe`, such as:

nwbfile.trials.to_dataframe()

####################
# This approach makes it easy to query the data to, e.g., locate all time intervals within a certain time range

trials_df = nwbfile.trials.to_dataframe()
trials_df.query("(start_time > 2.0) & (stop_time < 9.0)")

####################
# Accessing referenced TimeSeries
# -------------------------------
#
# As mentioned earlier, the ``timeseries`` column is defined by a :py:class:`~pynwb.base.TimeSeriesReferenceVectorData`
# which stores references to the corresponding ranges in :py:class:`~pynwb.base.TimeSeries`. Individual references
# to :py:class:`~pynwb.base.TimeSeries` are described via :py:class:`~pynwb.base.TimeSeriesReference` tuples
# with the :py:class:`~pynwb.base.TimeSeriesReference.idx_start`, :py:class:`~pynwb.base.TimeSeriesReference.count`,
# and :py:class:`~pynwb.base.TimeSeriesReference.timeseries`.
# Using :py:class:`~pynwb.base.TimeSeriesReference` we can easily access the relevant
# :py:meth:`~pynwb.base.TimeSeriesReference.data` and :py:meth:`~pynwb.base.TimeSeriesReference.timestamps`
# for the corresponding time range from the :py:class:`~pynwb.base.TimeSeries`.

# Get a single example TimeSeriesReference from the trials table
example_tsr = nwbfile.trials["timeseries"][0][0]

# Get the data values from the timeseries. This is a shorthand for:
# _ = example_tsr.timeseries.data[example_tsr.idx_start: (example_tsr.idx_start + example_tsr.count)]
_ = example_tsr.data

# Get the timestamps. Timestamps are either loaded from the TimeSeries or
# computed from the starting_time and rate
example_tsr.timestamps

####################
# Using :py:class:`~pynwb.base.TimeSeriesReference.isvalid` we can further check if the reference is valid.
# A :py:class:`~pynwb.base.TimeSeriesReference` is defined as invalid if both
# :py:class:`~pynwb.base.TimeSeriesReference.idx_start`, :py:class:`~pynwb.base.TimeSeriesReference.count` are
# set to ``-1``. :py:class:`~pynwb.base.TimeSeriesReference.isvalid` further also checks that the indicated
# index range and types are valid, raising ``IndexError`` and ``TypeError`` respectively, if bad
# :py:class:`~pynwb.base.TimeSeriesReference.idx_start`, :py:class:`~pynwb.base.TimeSeriesReference.count` or
# :py:class:`~pynwb.base.TimeSeriesReference.timeseries` are found.

example_tsr.isvalid()

####################
# Adding TimeSeries references to other tables
# --------------------------------------------
#
# Since :py:class:`~pynwb.base.TimeSeriesReferenceVectorData` is a regular :py:class:`~hdmf.common.table.VectorData`
# type, we can use it to add references to intervals in :py:class:`~pynwb.base.TimeSeries` to any
# :py:class:`~hdmf.common.table.DynamicTable`. In the :py:class:`~pynwb.icephys.IntracellularRecordingsTable`, e.g.,
# it is used to reference the recording of the stimulus and response associated with a particular intracellular
# electrophysiology recording.
#


####################
# Reading/Writing TimeIntervals to file
# -------------------------------------
#
# Reading and writing the data is as usual:

from pynwb import NWBHDF5IO

# write the file
with NWBHDF5IO("example_timeintervals_file.nwb", "w") as io:
    io.write(nwbfile)
# read the file
with NWBHDF5IO("example_timeintervals_file.nwb", "r") as io:
    nwbfile_in = io.read()

    # plot the sleep stages TimeIntervals table
    nwbfile_in.get_time_intervals("sleep_stages").to_dataframe()



"""
Module with helper functions to facilitate testing
"""
import numpy as np
from datetime import datetime
from dateutil.tz import tzlocal
from pynwb.file import NWBFile
from pynwb.icephys import (VoltageClampStimulusSeries, VoltageClampSeries)
from pynwb import NWBHDF5IO


def create_icephys_stimulus_and_response(sweep_number, electrode, randomize_data):
    """
    Internal helper function to construct a dummy stimulus and response pair representing an
    intracellular recording:

    :param sweep_number: Integer sweep number of the recording
    :type sweep_number: int
    :param electrode: Intracellular electrode used
    :type electrode: pynwb.icephys.IntracellularElectrode
    :param randomize_data: Randomize data values in the stimulus and response
    :type randomize_data: bool

    :returns: Tuple of VoltageClampStimulusSeries with the stimulus and VoltageClampSeries with the response.
    """
    stimulus = VoltageClampStimulusSeries(
                name="ccss_"+str(sweep_number),
                data=[1, 2, 3, 4, 5] if not randomize_data else np.random.rand(10),
                starting_time=123.6 if not randomize_data else (np.random.rand() * 100),
                rate=10e3 if not randomize_data else int(np.random.rand()*10) * 1000 + 1000.,
                electrode=electrode,
                gain=0.1 if not randomize_data else np.random.rand(),
                sweep_number=sweep_number)
    # Create and ic-response
    response = VoltageClampSeries(
                name='vcs_'+str(sweep_number),
                data=[0.1, 0.2, 0.3, 0.4, 0.5] if not randomize_data else np.random.rand(10),
                conversion=1e-12,
                resolution=np.nan,
                starting_time=123.6 if not randomize_data else (np.random.rand() * 100),
                rate=20e3 if not randomize_data else int(np.random.rand() * 20) * 1000. + 1000.,
                electrode=electrode,
                gain=0.02 if not randomize_data else np.random.rand(),
                capacitance_slow=100e-12,
                resistance_comp_correction=70.0 if not randomize_data else 70.0 + np.random.rand(),
                sweep_number=sweep_number)
    return stimulus, response


def create_icephys_testfile(filename=None, add_custom_columns=True, randomize_data=True, with_missing_stimulus=True):
    """
    Create a small but relatively complex icephys test file that
    we can use for testing of queries.

    :param filename: The name of the output file to be generated. If set to None then the file is not written
                     but only created in memory
    :type filename: str, None
    :param add_custom_colums: Add custom metadata columns to each table
    :type add_custom_colums: bool
    :param randomize_data: Randomize data values in the stimulus and response
    :type randomize_data: bool

    :returns: NWBFile object with icephys data created for writing. NOTE: If filename is provided then
              the file is written to disk, but the function does not read the file back. If
              you want to use the file from disk then you will need to read it with NWBHDF5IO.
    :rtype: NWBFile
    """
    nwbfile = NWBFile(
            session_description='my first synthetic recording',
            identifier='EXAMPLE_ID',
            session_start_time=datetime.now(tzlocal()),
            experimenter='Dr. Bilbo Baggins',
            lab='Bag End Laboratory',
            institution='University of Middle Earth at the Shire',
            experiment_description='I went on an adventure with thirteen dwarves to reclaim vast treasures.',
            session_id='LONELYMTN')
    # Add a device
    device = nwbfile.create_device(name='Heka ITC-1600')
    # Add an intracellular electrode
    electrode0 = nwbfile.create_icephys_electrode(
        name="elec0",
        description='a mock intracellular electrode',
        device=device)
    # Add an intracellular electrode
    electrode1 = nwbfile.create_icephys_electrode(
        name="elec1",
        description='another mock intracellular electrode',
        device=device)
    # Add the intracelluar recordings
    for sweep_number in range(20):
        elec = (electrode0 if (sweep_number % 2 == 0) else electrode1)
        stim, resp = create_icephys_stimulus_and_response(sweep_number=np.uint64(sweep_number),
                                                          electrode=elec,
                                                          randomize_data=randomize_data)
        if with_missing_stimulus and sweep_number in [0, 10]:
            stim = None
        nwbfile.add_intracellular_recording(electrode=elec,
                                            stimulus=stim,
                                            response=resp,
                                            id=sweep_number)
    nwbfile.intracellular_recordings.add_column(name='recording_tags',
                                                data=['A1', 'A2',
                                                      'B1', 'B2',
                                                      'C1', 'C2', 'C3',
                                                      'D1', 'D2', 'D3',
                                                      'A1', 'A2',
                                                      'B1', 'B2',
                                                      'C1', 'C2', 'C3',
                                                      'D1', 'D2', 'D3'],
                                                description='String with a set of recording tags')
    # Add simultaneous_recordings
    nwbfile.add_icephys_simultaneous_recording(recordings=[0, 1], id=np.int64(100))
    nwbfile.add_icephys_simultaneous_recording(recordings=[2, 3], id=np.int64(101))
    nwbfile.add_icephys_simultaneous_recording(recordings=[4, 5, 6], id=np.int64(102))
    nwbfile.add_icephys_simultaneous_recording(recordings=[7, 8, 9], id=np.int64(103))
    nwbfile.add_icephys_simultaneous_recording(recordings=[10, 11], id=np.int64(104))
    nwbfile.add_icephys_simultaneous_recording(recordings=[12, 13], id=np.int64(105))
    nwbfile.add_icephys_simultaneous_recording(recordings=[14, 15, 16], id=np.int64(106))
    nwbfile.add_icephys_simultaneous_recording(recordings=[17, 18, 19], id=np.int64(107))
    if add_custom_columns:
        nwbfile.icephys_simultaneous_recordings.add_column(
            name='tag',
            data=np.arange(8),
            description='some integer tag for a sweep')

    # Add sequential recordings
    nwbfile.add_icephys_sequential_recording(simultaneous_recordings=[0, 1],
                                             id=np.int64(1000),
                                             stimulus_type="StimType_1")
    nwbfile.add_icephys_sequential_recording(simultaneous_recordings=[2, ],
                                             id=np.int64(1001),
                                             stimulus_type="StimType_2")
    nwbfile.add_icephys_sequential_recording(simultaneous_recordings=[3, ],
                                             id=np.int64(1002),
                                             stimulus_type="StimType_3")
    nwbfile.add_icephys_sequential_recording(simultaneous_recordings=[4, 5],
                                             id=np.int64(1003),
                                             stimulus_type="StimType_1")
    nwbfile.add_icephys_sequential_recording(simultaneous_recordings=[6, ],
                                             id=np.int64(1004),
                                             stimulus_type="StimType_2")
    nwbfile.add_icephys_sequential_recording(simultaneous_recordings=[7, ],
                                             id=np.int64(1005),
                                             stimulus_type="StimType_3")
    if add_custom_columns:
        nwbfile.icephys_sequential_recordings.add_column(
            name='type',
            data=['T1', 'T2', 'T3', 'T1', 'T2', 'T3'],
            description='type of the sequential recording')

    # Add repetitions
    nwbfile.add_icephys_repetition(sequential_recordings=[0, ], id=np.int64(10000))
    nwbfile.add_icephys_repetition(sequential_recordings=[1, 2], id=np.int64(10001))
    nwbfile.add_icephys_repetition(sequential_recordings=[3, ], id=np.int64(10002))
    nwbfile.add_icephys_repetition(sequential_recordings=[4, 5], id=np.int64(10003))
    if add_custom_columns:
        nwbfile.icephys_repetitions.add_column(
            name='type',
            data=['R1', 'R2', 'R1', 'R2'],
            description='some repetition type indicator')

    # Add experimental_conditions
    nwbfile.add_icephys_experimental_condition(repetitions=[0, 1], id=np.int64(100000))
    nwbfile.add_icephys_experimental_condition(repetitions=[2, 3], id=np.int64(100001))
    if add_custom_columns:
        nwbfile.icephys_experimental_conditions.add_column(
            name='temperature',
            data=[32., 24.],
            description='Temperatur in C')

    # Write our test file
    if filename is not None:
        with NWBHDF5IO(filename, 'w') as io:
            io.write(nwbfile)

    # Return our in-memory NWBFile
    return nwbfile



from pynwb.epoch import TimeIntervals
from pynwb.testing import NWBH5IOMixin, TestCase
from pynwb import NWBFile, NWBHDF5IO
from pynwb.base import TimeSeries, TimeSeriesReference, TimeSeriesReferenceVectorData
import numpy as np
import h5py


class TestTimeIntervalsIO(NWBH5IOMixin, TestCase):
    """
    Test that reading NWB files with version 2.4 and earlier function correctly with the TimeIntervals type.
    In NWB 2.5 TimeIntervals.timeseries was updated to use the TimeSeriesReferenceVectorData type and to
    keep the API consistent old files are migrated on read to present the same API for those files. This
    is possible because the previous and updated schema are identical aside from the fact that a dedicated
    neurodata_type was created for the TimeIntervals.timeseries columns which is also used in the
    IntracellularRecordingTable type.

    This is technically a round-trip test, but without round-trip we can't test the object-mapper and
    including it here makes sure this gets included in code-coverage tests as well.
    """

    def setUpContainer(self):
        """ Return placeholder epochs object. Tested epochs are added directly to the NWBFile in addContainer """
        return TimeIntervals('epochs')

    def addContainer(self, nwbfile):
        """ Add the test epochs to the given NWBFile """
        # Add some timeseries
        tsa, tsb = [
            TimeSeries(name='a', data=np.arange(11), unit='flubs', timestamps=np.linspace(0, 1.0, 11)),
            TimeSeries(name='b', data=np.arange(13), unit='flubs', timestamps=np.linspace(0.1, 2.0, 13)),
        ]
        nwbfile.add_acquisition(tsa)
        nwbfile.add_acquisition(tsb)
        # Add a custom column
        nwbfile.add_epoch_column(
            name='temperature',
            description='average temperature (c) during epoch'
        )
        # Add some epochs
        nwbfile.add_epoch(
            start_time=0.0,
            stop_time=0.5,
            timeseries=[tsa, tsb],
            tags='ambient',
            temperature=26.4,
        )
        nwbfile.add_epoch(
            start_time=1.3,
            stop_time=4.1,
            timeseries=[tsa, ],
            tags='ambient',
            temperature=26.4,
        )
        # reset the thing
        self.container = nwbfile.epochs

    def getContainer(self, nwbfile):
        """ Return the test epochs from the given NWBFile """
        return nwbfile.epochs

    def test_legacy_format(self):
        description = 'a file to test writing and reading a %s' % self.container_type
        identifier = 'TEST_%s' % self.container_type
        nwbfile = NWBFile(description, identifier, self.start_time, file_create_date=self.create_date)
        self.addContainer(nwbfile)

        # write the file
        with NWBHDF5IO(self.filename, mode='w') as write_io:
            write_io.write(nwbfile, cache_spec=False)
        # Modify the HDF5 file to look like NWB 2.4 and earlier. This simply means
        # modifying the neurodata_type on the TimeIntervals.timeseries column
        with h5py.File(self.filename, mode='a') as infile:
            infile['/intervals/epochs/timeseries'].attrs['neurodata_type'] = 'VectorData'
            infile.attrs['nwb_version'] = '2.3.0'

        # Read the file back
        self.reader = NWBHDF5IO(self.filename, mode='r')
        self.read_nwbfile = self.reader.read()

        # Test that the VectorData column for timeseries has been converted to TimeSeriesReferenceVectorData
        self.assertIsInstance(self.read_nwbfile.epochs.timeseries, TimeSeriesReferenceVectorData)

        # Test that slicing into epochs.timeseries works as expected
        re = self.read_nwbfile.epochs.timeseries[0]
        self.assertIsInstance(re, TimeSeriesReference)
        self.assertTupleEqual((re[0], re[1], re[2].object_id), (0, 5, nwbfile.get_acquisition('a').object_id))

        # Test that slicing into epochs works as expected
        re = self.read_nwbfile.epochs[0:1]
        self.assertListEqual(re.columns.tolist(), ['start_time', 'stop_time', 'temperature', 'tags', 'timeseries'])
        for i in re.loc[0, 'timeseries']:
            self.assertIsInstance(i, TimeSeriesReference)
        self.assertTupleEqual(
            (re.loc[0, 'timeseries'][0][0], re.loc[0, 'timeseries'][0][1], re.loc[0, 'timeseries'][0][2].object_id),
            (0, 5, nwbfile.get_acquisition('a').object_id))
        self.assertTupleEqual(
            (re.loc[0, 'timeseries'][1][0], re.loc[0, 'timeseries'][1][1], re.loc[0, 'timeseries'][1][2].object_id),
            (0, 3, nwbfile.get_acquisition('b').object_id))



from pynwb.testing import TestCase


class TestRetinotopy(TestCase):
    def test_retinotopy_deprecated(self):
        with self.assertRaises(ImportError):
            import pynwb.retinotopy  # noqa: F401

# import numpy as np

# from pynwb.retinotopy import ImagingRetinotopy, AxisMap, RetinotopyImage, FocalDepthImage, RetinotopyMap
# from pynwb.testing import TestCase


# class ImageRetinotopyConstructor(TestCase):

#     def setUp(self):
#         data = np.ones((2, 2))
#         field_of_view = [1, 2]
#         dimension = [1, 2]
#         self.sign_map = RetinotopyMap('sign_map', data, field_of_view, dimension)
#         self.axis_1_phase_map = AxisMap('axis_1_phase_map', data, field_of_view, 'unit', dimension)
#         self.axis_1_power_map = AxisMap('axis_1_power_map', data, field_of_view, 'unit', dimension)
#         self.axis_2_phase_map = AxisMap('axis_2_phase_map', data, field_of_view, 'unit', dimension)
#         self.axis_2_power_map = AxisMap('axis_2_power_map', data, field_of_view, 'unit', dimension)
#         self.axis_descriptions = ['altitude', 'azimuth']

#         data = [[1, 1], [1, 1]]
#         bits_per_pixel = 8
#         dimension = [3, 4]
#         format = 'raw'
#         field_of_view = [1, 2]
#         focal_depth = 1.0
#         self.focal_depth_image = FocalDepthImage('focal_depth_image', data, bits_per_pixel, dimension, format,
#                                                  field_of_view, focal_depth)
#         self.vasculature_image = RetinotopyImage('vasculature_image', np.uint16(data), bits_per_pixel, dimension,
#                                                  format, field_of_view)

#     def test_init(self):
#         """Test that ImagingRetinotopy constructor sets properties correctly."""
#         msg = ('The ImagingRetinotopy class currently cannot be written to or read from a file. This is a known bug '
#                'and will be fixed in a future release of PyNWB.')
#         with self.assertWarnsWith(UserWarning, msg):
#             ir = ImagingRetinotopy(self.sign_map, self.axis_1_phase_map, self.axis_1_power_map, self.axis_2_phase_map,
#                                    self.axis_2_power_map, self.axis_descriptions, self.focal_depth_image,
#                                    self.vasculature_image)
#         self.assertEqual(ir.sign_map, self.sign_map)
#         self.assertEqual(ir.axis_1_phase_map, self.axis_1_phase_map)
#         self.assertEqual(ir.axis_1_power_map, self.axis_1_power_map)
#         self.assertEqual(ir.axis_2_phase_map, self.axis_2_phase_map)
#         self.assertEqual(ir.axis_2_power_map, self.axis_2_power_map)
#         self.assertEqual(ir.axis_descriptions, self.axis_descriptions)
#         self.assertEqual(ir.focal_depth_image, self.focal_depth_image)
#         self.assertEqual(ir.vasculature_image, self.vasculature_image)

#     def test_init_axis_descriptions_wrong_shape(self):
#         """Test that creating a ImagingRetinotopy with a axis descriptions argument that is not 2 elements raises an
#         error.
#         """
#         self.axis_descriptions = ['altitude', 'azimuth', 'extra']

#         msg = "ImagingRetinotopy.__init__: incorrect shape for 'axis_descriptions' (got '(3,)', expected '(2,)')"
#         with self.assertRaisesWith(ValueError, msg):
#             ImagingRetinotopy(self.sign_map, self.axis_1_phase_map, self.axis_1_power_map, self.axis_2_phase_map,
#                               self.axis_2_power_map, self.axis_descriptions, self.focal_depth_image,
#                               self.vasculature_image)


# class RetinotopyImageConstructor(TestCase):

#     def test_init(self):
#         """Test that RetinotopyImage constructor sets properties correctly."""
#         data = [[1, 1], [1, 1]]
#         bits_per_pixel = 8
#         dimension = [3, 4]
#         format = 'raw'
#         field_of_view = [1, 2]
#         image = RetinotopyImage('vasculature_image', data, bits_per_pixel, dimension, format, field_of_view)

#         self.assertEqual(image.name, 'vasculature_image')
#         self.assertEqual(image.data, data)
#         self.assertEqual(image.bits_per_pixel, bits_per_pixel)
#         self.assertEqual(image.dimension, dimension)
#         self.assertEqual(image.format, format)
#         self.assertEqual(image.field_of_view, field_of_view)

#     def test_init_dimension_wrong_shape(self):
#         """Test that creating a RetinotopyImage with a dimension argument that is not 2 elements raises an error."""
#         data = [[1, 1], [1, 1]]
#         bits_per_pixel = 8
#         dimension = [3, 4, 5]
#         format = 'raw'
#         field_of_view = [1, 2]

#         msg = "RetinotopyImage.__init__: incorrect shape for 'dimension' (got '(3,)', expected '(2,)')"
#         with self.assertRaisesWith(ValueError, msg):
#             RetinotopyImage('vasculature_image', data, bits_per_pixel, dimension, format, field_of_view)

#     def test_init_fov_wrong_shape(self):
#         """Test that creating a RetinotopyImage with a field of view that is not 2 elements raises an error."""
#         data = [[1, 1], [1, 1]]
#         bits_per_pixel = 8
#         dimension = [3, 4]
#         format = 'raw'
#         field_of_view = [1, 2, 3]

#         msg = "RetinotopyImage.__init__: incorrect shape for 'field_of_view' (got '(3,)', expected '(2,)')"
#         with self.assertRaisesWith(ValueError, msg):
#             RetinotopyImage('vasculature_image', data, bits_per_pixel, dimension, format, field_of_view)


# class RetinotopyMapConstructor(TestCase):

#     def test_init(self):
#         """Test that RetinotopyMap constructor sets properties correctly."""
#         data = np.ones((2, 2))
#         field_of_view = [1, 2]
#         dimension = [1, 2]
#         map = RetinotopyMap('sign_map', data, field_of_view, dimension)

#         self.assertEqual(map.name, 'sign_map')
#         np.testing.assert_array_equal(map.data, data)
#         self.assertEqual(map.field_of_view, field_of_view)
#         self.assertEqual(map.dimension, dimension)


# class AxisMapConstructor(TestCase):

#     def test_init(self):
#         """Test that AxisMap constructor sets properties correctly."""
#         data = np.ones((2, 2))
#         field_of_view = [1, 2]
#         dimension = [1, 2]
#         map = AxisMap('axis_1_phase', data, field_of_view, 'unit', dimension)

#         self.assertEqual(map.name, 'axis_1_phase')
#         np.testing.assert_array_equal(map.data, data)
#         self.assertEqual(map.field_of_view, field_of_view)
#         self.assertEqual(map.dimension, dimension)
#         self.assertEqual(map.unit, 'unit')

#     def test_init_dimension_wrong_shape(self):
#         """Test that creating an AxisMap with a dimension argument that is not 2 elements raises an error."""
#         data = np.ones((2, 2))
#         field_of_view = [1, 2]
#         dimension = [1, 2, 3]

#         msg = "AxisMap.__init__: incorrect shape for 'dimension' (got '(3,)', expected '(2,)')"
#         with self.assertRaisesWith(ValueError, msg):
#             AxisMap('axis_1_phase', data, field_of_view, 'unit', dimension)

#     def test_init_fov_wrong_shape(self):
#         """Test that creating an AxisMap with a dimension argument that is not 2 elements raises an error."""
#         data = np.ones((2, 2))
#         field_of_view = [1, 2, 3]
#         dimension = [1, 2]

#         msg = "AxisMap.__init__: incorrect shape for 'field_of_view' (got '(3,)', expected '(2,)')"
#         with self.assertRaisesWith(ValueError, msg):
#             AxisMap('axis_1_phase', data, field_of_view, 'unit', dimension)


# class FocalDepthImageConstructor(TestCase):

#     def test_init(self):
#         """Test that FocalDepthImage constructor sets properties correctly."""
#         data = [[1, 1], [1, 1]]
#         bits_per_pixel = 8
#         dimension = [3, 4]
#         format = 'raw'
#         field_of_view = [1, 2]
#         focal_depth = 1.0
#         image = FocalDepthImage('focal_depth_image', data, bits_per_pixel, dimension, format, field_of_view,
#                                 focal_depth)

#         self.assertEqual(image.name, 'focal_depth_image')
#         self.assertEqual(image.data, data)
#         self.assertEqual(image.bits_per_pixel, bits_per_pixel)
#         self.assertEqual(image.dimension, dimension)
#         self.assertEqual(image.format, format)
#         self.assertEqual(image.field_of_view, field_of_view)
#         self.assertEqual(image.focal_depth, focal_depth)



name_generator_registry = dict()


def name_generator(name):
    """
    Returns unique names of neurodata types using an incrementing number. The first time you pass "TimeSeries" it
    returns "TimeSeries". The second time, it returns "TimeSeries2", the third time it returns "TimeSeries3", etc.

    Parameters
    ----------
    name: str
        name of neurodata_type, e.g. TimeSeries

    Returns
    -------
    name of neurodata_object: str
    """
    if name not in name_generator_registry:
        name_generator_registry[name] = 1
        return name
    else:
        name_generator_registry[name] += 1
        return f"{name}{name_generator_registry[name]}"



# -*- coding: utf-8 -*-
"""
.. _icephys_tutorial:

Intracellular Electrophysiology Data using SweepTable
=====================================================

The following tutorial describes storage of intracellular electrophysiology data in NWB using the
SweepTable to manage recordings.

.. warning::
    The use of SweepTable has been deprecated as of PyNWB >v2.0 in favor of the new hierarchical
    intracellular electrophysiology metadata tables to allow for a more complete description of
    intracellular electrophysiology experiments. See the :doc:`Intracellular electrophysiology  <plot_icephys>`
    tutorial for details.
"""

#######################
# Creating and Writing NWB files
# ------------------------------
#
# When creating a NWB file, the first step is to create the :py:class:`~pynwb.file.NWBFile`. The first
# argument is is a brief description of the dataset.

# sphinx_gallery_thumbnail_path = 'figures/gallery_thumbnails_icephys_sweeptable.png'
from datetime import datetime
from uuid import uuid4

import numpy as np
from dateutil.tz import tzlocal

from pynwb import NWBFile

nwbfile = NWBFile(
    session_description="my first synthetic recording",
    identifier=str(uuid4()),
    session_start_time=datetime.now(tzlocal()),
    experimenter=[
        "Baggins, Bilbo",
    ],
    lab="Bag End Laboratory",
    institution="University of Middle Earth at the Shire",
    experiment_description="I went on an adventure to reclaim vast treasures.",
    session_id="LONELYMTN001",
)

#######################
# Device metadata
# ^^^^^^^^^^^^^^^
#
# Device metadata is represented by :py:class:`~pynwb.device.Device` objects.
# To create a device, you can use the :py:class:`~pynwb.device.Device` instance method
# :py:meth:`~pynwb.file.NWBFile.create_device`.

device = nwbfile.create_device(name="Heka ITC-1600")

#######################
# Electrode metadata
# ^^^^^^^^^^^^^^^^^^
#
# Intracellular electrode metadata is represented by :py:class:`~pynwb.icephys.IntracellularElectrode` objects.
# To create an electrode group, you can use the :py:class:`~pynwb.file.NWBFile` instance method
# :py:meth:`~pynwb.file.NWBFile.create_icephys_electrode`.

elec = nwbfile.create_icephys_electrode(
    name="elec0", description="a mock intracellular electrode", device=device
)

#######################
# Stimulus data
# ^^^^^^^^^^^^^
#
# Intracellular stimulus and response data are represented with subclasses of
# :py:class:`~pynwb.icephys.PatchClampSeries`. There are two classes for representing stimulus
# data
#
# - :py:class:`~pynwb.icephys.VoltageClampStimulusSeries`
# - :py:class:`~pynwb.icephys.CurrentClampStimulusSeries`
#
# and three classes for representing response
#
# - :py:class:`~pynwb.icephys.VoltageClampSeries`
# - :py:class:`~pynwb.icephys.CurrentClampSeries`
# - :py:class:`~pynwb.icephys.IZeroClampSeries`
#
# Here, we will use :py:class:`~pynwb.icephys.CurrentClampStimulusSeries` to store current clamp stimulus
# data and then add it to our NWBFile as stimulus data using the :py:class:`~pynwb.file.NWBFile` method
# :py:meth:`~pynwb.file.NWBFile.add_stimulus`.

from pynwb.icephys import CurrentClampStimulusSeries

ccss = CurrentClampStimulusSeries(
    name="ccss",
    data=[1, 2, 3, 4, 5],
    starting_time=123.6,
    rate=10e3,
    electrode=elec,
    gain=0.02,
    sweep_number=0,
)

nwbfile.add_stimulus(ccss, use_sweep_table=True)

#######################
# We now add another stimulus series but from a different sweep. TimeSeries
# having the same starting time belong to the same sweep.

from pynwb.icephys import VoltageClampStimulusSeries

vcss = VoltageClampStimulusSeries(
    name="vcss",
    data=[2, 3, 4, 5, 6],
    starting_time=234.5,
    rate=10e3,
    electrode=elec,
    gain=0.03,
    sweep_number=1,
)

nwbfile.add_stimulus(vcss, use_sweep_table=True)

#######################
# Here, we will use :py:class:`~pynwb.icephys.CurrentClampSeries` to store current clamp
# data and then add it to our NWBFile as acquired data using the :py:class:`~pynwb.file.NWBFile` method
# :py:meth:`~pynwb.file.NWBFile.add_acquisition`.

from pynwb.icephys import CurrentClampSeries

ccs = CurrentClampSeries(
    name="ccs",
    data=[0.1, 0.2, 0.3, 0.4, 0.5],
    conversion=1e-12,
    resolution=np.nan,
    starting_time=123.6,
    rate=20e3,
    electrode=elec,
    gain=0.02,
    bias_current=1e-12,
    bridge_balance=70e6,
    capacitance_compensation=1e-12,
    sweep_number=0,
)

nwbfile.add_acquisition(ccs, use_sweep_table=True)

#######################
# And voltage clamp data from the second sweep using
# :py:class:`~pynwb.icephys.VoltageClampSeries`.

from pynwb.icephys import VoltageClampSeries

vcs = VoltageClampSeries(
    name="vcs",
    data=[0.1, 0.2, 0.3, 0.4, 0.5],
    conversion=1e-12,
    resolution=np.nan,
    starting_time=234.5,
    rate=20e3,
    electrode=elec,
    gain=0.02,
    capacitance_slow=100e-12,
    resistance_comp_correction=70.0,
    sweep_number=1,
)

nwbfile.add_acquisition(vcs, use_sweep_table=True)

####################
# .. _icephys_writing:
#
# Once you have finished adding all of your data to the :py:class:`~pynwb.file.NWBFile`,
# write the file with :py:class:`~pynwb.NWBHDF5IO`.

from pynwb import NWBHDF5IO

with NWBHDF5IO("icephys_example.nwb", "w") as io:
    io.write(nwbfile)

####################
# For more details on :py:class:`~pynwb.NWBHDF5IO`, see the :ref:`basic tutorial <basic_writing>`.

####################
# .. _icephys_reading:
#
# Reading electrophysiology data
# ------------------------------
#
# Now that you have written some intracellular electrophysiology data, you can read it back in.

io = NWBHDF5IO("icephys_example.nwb", "r")
nwbfile = io.read()

####################
# For details on retrieving data from an :py:class:`~pynwb.file.NWBFile`, we refer the reader to the
# :ref:`basic tutorial <basic_reading>`. For this tutorial, we will just get back our the
# :py:class:`~pynwb.icephys.CurrentClampStimulusSeries` object we added above.
#
# First, get the :py:class:`~pynwb.icephys.CurrentClampStimulusSeries` we added as stimulus data.

ccss = nwbfile.get_stimulus("ccss")

####################
# Grabbing acquisition data can be done via :py:meth:`~pynwb.file.NWBFile.get_acquisition`

vcs = nwbfile.get_acquisition("vcs")

####################
# We can also get back the electrode we added.

elec = nwbfile.get_icephys_electrode("elec0")

####################
# Alternatively, we can also get this electrode from the :py:class:`~pynwb.icephys.CurrentClampStimulusSeries`
# we retrieved above. This is exposed via the :py:meth:`~pynwb.icephys.PatchClampSeries.electrode` attribute.

elec = ccss.electrode

####################
# And the device name via :py:meth:`~pynwb.file.NWBFile.get_device`

device = nwbfile.get_device("Heka ITC-1600")

####################
# If you have data from multiple electrodes and multiple sweeps, it can be
# tedious and expensive to search all :py:class:`~pynwb.icephys.PatchClampSeries` for the
# :py:class:`~pynwb.base.TimeSeries` with a given sweep.
#
# Fortunately you don't have to do that manually, instead you can just query
# the :py:class:`~pynwb.icephys.SweepTable` which stores the mapping between the
# PatchClampSeries which belongs to a certain sweep number via
# :py:meth:`~pynwb.icephys.SweepTable.get_series`.
#
# The following call will return the voltage clamp data of two timeseries
# consisting of acquisition and stimulus, from sweep 1.

series = nwbfile.sweep_table.get_series(1)

# close the file
io.close()



"""
Parallel I/O using MPI
======================

The HDF5 storage backend supports parallel I/O using the Message Passing Interface (MPI).
Using this feature requires that you install ``hdf5`` and ``h5py`` against an MPI driver, and you
install ``mpi4py``. The basic installation of pynwb will not work. Setup can be tricky, and
is outside the scope of this tutorial (for now), and the following assumes that you have
HDF5 installed in a MPI configuration.
"""

# sphinx_gallery_thumbnail_path = 'figures/gallery_thumbnails_parallelio.png'

####################
# Here we:
#
# 1. **Instantiate a dataset for parallel write**: We create TimeSeries with 4 timestamps that we
# will write in parallel
#
# 2. **Write to that file in parallel using MPI**: Here we assume 4 MPI ranks while each rank writes
# the data for a different timestamp.
#
# 3. **Read from the file in parallel using MPI**: Here each of the 4 MPI ranks reads one time
# step from the file
#
# .. code-block:: python
#
#     from mpi4py import MPI
#     import numpy as np
#     from dateutil import tz
#     from pynwb import NWBHDF5IO, NWBFile, TimeSeries
#     from datetime import datetime
#     from hdmf.backends.hdf5.h5_utils import H5DataIO
#
#     start_time = datetime(2018, 4, 25, 2, 30, 3, tzinfo=tz.gettz("US/Pacific"))
#     fname = "test_parallel_pynwb.nwb"
#     rank = MPI.COMM_WORLD.rank  # The process ID (integer 0-3 for 4-process run)
#
#     # Create file on one rank. Here we only instantiate the dataset we want to
#     # write in parallel but we do not write any data
#     if rank == 0:
#         nwbfile = NWBFile("aa", "aa", start_time)
#         data = H5DataIO(shape=(4,), maxshape=(4,), dtype=np.dtype("int"))
#
#         nwbfile.add_acquisition(
#             TimeSeries(name="ts_name", description="desc", data=data, rate=100.0, unit="m")
#         )
#         with NWBHDF5IO(fname, "w") as io:
#             io.write(nwbfile)
#
#     # write to dataset in parallel
#     with NWBHDF5IO(fname, "a", comm=MPI.COMM_WORLD) as io:
#         nwbfile = io.read()
#         print(rank)
#         nwbfile.acquisition["ts_name"].data[rank] = rank
#
#     # read from dataset in parallel
#     with NWBHDF5IO(fname, "r", comm=MPI.COMM_WORLD) as io:
#         print(io.read().acquisition["ts_name"].data[rank])

####################
#
# .. note::
#
#    Using :py:class:`hdmf.backends.hdf5.h5_utils.H5DataIO` we can also specify further
#    details about the data layout, e.g., via the chunking and compression parameters.
#



'''This package will contain functions, classes, and objects
for reading and writing data in NWB format
'''
import os.path
from pathlib import Path
from copy import deepcopy
from warnings import warn
import h5py

from hdmf.spec import NamespaceCatalog
from hdmf.utils import docval, getargs, popargs, get_docval
from hdmf.backends.io import HDMFIO
from hdmf.backends.hdf5 import HDF5IO as _HDF5IO
from hdmf.build import BuildManager, TypeMap
import hdmf.common
from hdmf.common import load_type_config as hdmf_load_type_config
from hdmf.common import get_loaded_type_config as hdmf_get_loaded_type_config
from hdmf.common import unload_type_config as hdmf_unload_type_config


CORE_NAMESPACE = 'core'

from .spec import NWBDatasetSpec, NWBGroupSpec, NWBNamespace  # noqa E402
from .validate import validate  # noqa: F401, E402


@docval({'name': 'config_path', 'type': str, 'doc': 'Path to the configuration file.'},
        {'name': 'type_map', 'type': TypeMap, 'doc': 'The TypeMap.', 'default': None},
        is_method=False)
def load_type_config(**kwargs):
    """
    This method will either load the default config or the config provided by the path.
    """
    config_path = kwargs['config_path']
    type_map = kwargs['type_map'] or get_type_map()

    hdmf_load_type_config(config_path=config_path, type_map=type_map)

@docval({'name': 'type_map', 'type': TypeMap, 'doc': 'The TypeMap.', 'default': None},
        is_method=False)
def get_loaded_type_config(**kwargs):
    type_map = kwargs['type_map'] or get_type_map()
    return hdmf_get_loaded_type_config(type_map=type_map)

@docval({'name': 'type_map', 'type': TypeMap, 'doc': 'The TypeMap.', 'default': None},
        is_method=False)
def unload_type_config(**kwargs):
    """
    Remove validation.
    """
    type_map = kwargs['type_map'] or get_type_map()
    hdmf_unload_type_config(type_map=type_map)

def __get_resources():
    try:
        from importlib.resources import files
    except ImportError:
        # TODO: Remove when python 3.9 becomes the new minimum
        from importlib_resources import files

    __location_of_this_file = files(__name__)
    __core_ns_file_name = 'nwb.namespace.yaml'
    __schema_dir = 'nwb-schema/core'

    ret = dict()
    ret['namespace_path'] = str(__location_of_this_file / __schema_dir / __core_ns_file_name)
    return ret


def _get_resources():
    # LEGACY: Needed to support legacy implementation.
    return __get_resources()


# a global namespace catalog
global __NS_CATALOG
global __TYPE_MAP

__NS_CATALOG = NamespaceCatalog(NWBGroupSpec, NWBDatasetSpec, NWBNamespace)

hdmf_typemap = hdmf.common.get_type_map()
__TYPE_MAP = TypeMap(__NS_CATALOG)
__TYPE_MAP.merge(hdmf_typemap, ns_catalog=True)


@docval({'name': 'extensions', 'type': (str, TypeMap, list),
         'doc': 'a path to a namespace, a TypeMap, or a list consisting of paths to namespaces and TypeMaps',
         'default': None},
        returns="TypeMap loaded for the given extension or NWB core namespace", rtype=tuple,
        is_method=False)
def get_type_map(**kwargs):
    '''
    Get the TypeMap for the given extensions. If no extensions are provided,
    return the TypeMap for the core namespace
    '''
    extensions = getargs('extensions', kwargs)
    type_map = None
    if extensions is None:
        type_map = deepcopy(__TYPE_MAP)
    else:
        if isinstance(extensions, TypeMap):
            type_map = extensions
        else:
            type_map = deepcopy(__TYPE_MAP)
        if isinstance(extensions, list):
            for ext in extensions:
                if isinstance(ext, str):
                    type_map.load_namespaces(ext)
                elif isinstance(ext, TypeMap):
                    type_map.merge(ext)
                else:
                    raise ValueError('extensions must be a list of paths to namespace specs or a TypeMaps')
        elif isinstance(extensions, str):
            type_map.load_namespaces(extensions)
        elif isinstance(extensions, TypeMap):
            type_map.merge(extensions)
    return type_map


@docval(*get_docval(get_type_map),
        returns="the namespaces loaded from the given file", rtype=tuple,
        is_method=False)
def get_manager(**kwargs):
    '''
    Get a BuildManager to use for I/O using the given extensions. If no extensions are provided,
    return a BuildManager that uses the core namespace
    '''
    type_map = get_type_map(**kwargs)
    return BuildManager(type_map)


@docval({'name': 'namespace_path', 'type': str, 'doc': 'the path to the YAML with the namespace definition'},
        returns="the namespaces loaded from the given file", rtype=tuple,
        is_method=False)
def load_namespaces(**kwargs):
    '''
    Load namespaces from file
    '''
    namespace_path = getargs('namespace_path', kwargs)
    return __TYPE_MAP.load_namespaces(namespace_path)


# load the core namespace, i.e. base NWB specification
__resources = __get_resources()
if os.path.exists(__resources['namespace_path']):
    load_namespaces(__resources['namespace_path'])
else:
    raise RuntimeError(
        "'core' is not a registered namespace. If you installed PyNWB locally using a git clone, you need to "
        "use the --recurse_submodules flag when cloning. See developer installation instructions here: "
        "https://pynwb.readthedocs.io/en/stable/install_developers.html#install-from-git-repository"
    )


def available_namespaces():
    """Returns all namespaces registered in the namespace catalog"""
    return __NS_CATALOG.namespaces


# a function to register a container classes with the global map
@docval({'name': 'neurodata_type', 'type': str, 'doc': 'the neurodata_type to get the spec for'},
        {'name': 'namespace', 'type': str, 'doc': 'the name of the namespace'},
        {"name": "container_cls", "type": type, "doc": "the class to map to the specified neurodata_type",
         'default': None},
        is_method=False)
def register_class(**kwargs):
    """Register an NWBContainer class to use for reading and writing a neurodata_type from a specification
    If container_cls is not specified, returns a decorator for registering an NWBContainer subclass
    as the class for neurodata_type in namespace.
    """
    neurodata_type, namespace, container_cls = getargs('neurodata_type', 'namespace', 'container_cls', kwargs)

    def _dec(cls):
        __TYPE_MAP.register_container_type(namespace, neurodata_type, cls)
        return cls
    if container_cls is None:
        return _dec
    else:
        _dec(container_cls)


@docval({'name': 'h5py_file', 'type': h5py.File, 'doc': 'An NWB file'}, rtype=tuple,
        is_method=False,)
def get_nwbfile_version(**kwargs):
    """
    Get the NWB version of the file if it is an NWB file.

    :Returns: Tuple consisting of: 1) the
        original version string as stored in the file and 2) a tuple with the parsed components of the version string,
        consisting of integers and strings, e.g., (2, 5, 1, beta). (None, None) will be returned if the file is not a
        valid NWB file or the nwb_version is missing, e.g., in the case when no data has been written to the file yet.
    """
    # Get the version string for the NWB file
    h5py_file = getargs('h5py_file', kwargs)
    try:
        nwb_version_string = h5py_file.attrs['nwb_version']
    #  KeyError occurs  when the file is empty (e.g., when creating a new file nothing has been written)
    #  or when the HDF5 file is not a valid NWB file
    except KeyError:
        return None, None
    # Other system may have written nwb_version as a fixed-length string, resulting in a numpy.bytes_ object
    # on read, rather than a variable-length string. To address this, decode the bytes if necessary.
    if not isinstance(nwb_version_string, str):
        nwb_version_string = nwb_version_string.decode()

    # Parse the version string
    nwb_version_parts = nwb_version_string.replace("-", ".").replace("_", ".").split(".")
    nwb_version = tuple([int(i) if i.isnumeric() else i
                         for i in nwb_version_parts])
    return nwb_version_string, nwb_version


# a function to register an object mapper for a container class
@docval({"name": "container_cls", "type": type,
         "doc": "the Container class for which the given ObjectMapper class gets used"},
        {"name": "mapper_cls", "type": type, "doc": "the ObjectMapper class to use to map", 'default': None},
        is_method=False)
def register_map(**kwargs):
    """Register an ObjectMapper to use for a Container class type
    If mapper_cls is not specified, returns a decorator for registering an ObjectMapper class as the mapper for
    container_cls. If mapper_cls is specified, register the class as the mapper for container_cls
    """
    container_cls, mapper_cls = getargs('container_cls', 'mapper_cls', kwargs)

    def _dec(cls):
        __TYPE_MAP.register_map(container_cls, cls)
        return cls
    if mapper_cls is None:
        return _dec
    else:
        _dec(mapper_cls)


@docval({'name': 'neurodata_type', 'type': str, 'doc': 'the neurodata_type to get the NWBContainer class for'},
        {'name': 'namespace', 'type': str, 'doc': 'the namespace the neurodata_type is defined in'},
        is_method=False)
def get_class(**kwargs):
    """
    Parse the YAML file for a given neurodata_type that is a subclass of NWBContainer and automatically generate its
    python API. This will work for most containers, but is known to not work for descendants of MultiContainerInterface
    and DynamicTable, so these must be defined manually (for now). `get_class` infers the API mapping directly from the
    specification. If you want to define a custom mapping, you should not use this function and you should define the
    class manually.

    Examples:

    Generating and registering an extension is as simple as::

        MyClass = get_class('MyClass', 'ndx-my-extension')

    `get_class` defines only the `__init__` for the class. In cases where you want to provide additional methods for
    querying, plotting, etc. you can still use `get_class` and attach methods to the class after-the-fact, e.g.::

        def get_sum(self, a, b):
            return self.feat1 + self.feat2

        MyClass.get_sum = get_sum

    """
    neurodata_type, namespace = getargs('neurodata_type', 'namespace', kwargs)
    return __TYPE_MAP.get_dt_container_cls(neurodata_type, namespace)


class NWBHDF5IO(_HDF5IO):

    @staticmethod
    def can_read(path: str):
        """Determine whether a given path is readable by this class"""
        if not os.path.isfile(path):  # path is file that exists
            return False
        try:
            with h5py.File(path, "r") as file:   # path is HDF5 file
                version_info = get_nwbfile_version(file)
                if version_info[0] is None:
                    warn("Cannot read because missing NWB version in the HDF5 file. The file is not a valid NWB file.")
                    return False
                elif version_info[1][0] < 2:    # Major versions of NWB < 2 not supported
                    warn("Cannot read because PyNWB supports NWB files version 2 and above.")
                    return False
                else:
                    return True
        except IOError:
            return False

    @docval({'name': 'path', 'type': (str, Path), 'doc': 'the path to the HDF5 file', 'default': None},
            {'name': 'mode', 'type': str,
             'doc': 'the mode to open the HDF5 file with, one of ("w", "r", "r+", "a", "w-", "x")',
             'default': 'r'},
            {'name': 'load_namespaces', 'type': bool,
             'doc': ('whether or not to load cached namespaces from given path - not applicable in write mode '
                     'or when `manager` is not None or when `extensions` is not None'),
             'default': True},
            {'name': 'manager', 'type': BuildManager, 'doc': 'the BuildManager to use for I/O', 'default': None},
            {'name': 'extensions', 'type': (str, TypeMap, list),
             'doc': 'a path to a namespace, a TypeMap, or a list consisting paths to namespaces and TypeMaps',
             'default': None},
            *get_docval(_HDF5IO.__init__, "file", "comm", "driver", "aws_region", "herd_path"),)
    def __init__(self, **kwargs):
        path, mode, manager, extensions, load_namespaces, file_obj, comm, driver, aws_region, herd_path =\
            popargs('path', 'mode', 'manager', 'extensions', 'load_namespaces',
                    'file', 'comm', 'driver', 'aws_region', 'herd_path', kwargs)
        # Define the BuildManager to use
        io_modes_that_create_file = ['w', 'w-', 'x']
        if mode in io_modes_that_create_file or manager is not None or extensions is not None:
            load_namespaces = False

        if load_namespaces:
            tm = get_type_map()
            super().load_namespaces(tm, path, file=file_obj, driver=driver, aws_region=aws_region)
            manager = BuildManager(tm)

            # XXX: Leaving this here in case we want to revert to this strategy for
            #      loading cached namespaces
            # ns_catalog = NamespaceCatalog(NWBGroupSpec, NWBDatasetSpec, NWBNamespace)
            # super().load_namespaces(ns_catalog, path)
            # tm = TypeMap(ns_catalog)
            # tm.copy_mappers(get_type_map())
        else:
            if manager is not None and extensions is not None:
                raise ValueError("'manager' and 'extensions' cannot be specified together")
            elif extensions is not None:
                manager = get_manager(extensions=extensions)
            elif manager is None:
                manager = get_manager()
        # Open the file
        super().__init__(path, manager=manager, mode=mode, file=file_obj, comm=comm,
                         driver=driver, aws_region=aws_region, herd_path=herd_path)

    @property
    def nwb_version(self):
        """
        Get the version of the NWB file opened via this NWBHDF5IO object.

        :returns: Tuple consisting of: 1) the original version string as stored in the file and
                  2) a tuple with the parsed components of the version string, consisting of integers
                  and strings, e.g., (2, 5, 1, beta). (None, None) will be returned if the nwb_version
                  is missing, e.g., in the case when no data has been written to the file yet.
        """
        return get_nwbfile_version(self._file)

    @docval(*get_docval(_HDF5IO.read),
            {'name': 'skip_version_check', 'type': bool, 'doc': 'skip checking of NWB version', 'default': False})
    def read(self, **kwargs):
        """
        Read the NWB file from the IO source.

        :raises TypeError: If the NWB file version is missing or not supported

        :return: NWBFile container
        """
        # Check that the NWB file is supported
        skip_verison_check = popargs('skip_version_check', kwargs)
        if not skip_verison_check:
            file_version_str, file_version = self.nwb_version
            if file_version is None:
                raise TypeError("Missing NWB version in file. The file is not a valid NWB file.")
            if file_version[0] < 2:
                raise TypeError("NWB version %s not supported. PyNWB supports NWB files version 2 and above." %
                                str(file_version_str))
        # read the file
        file = super().read(**kwargs)
        return file

    @docval({'name': 'src_io', 'type': HDMFIO,
             'doc': 'the HDMFIO object (such as NWBHDF5IO) that was used to read the data to export'},
            {'name': 'nwbfile', 'type': 'NWBFile',
             'doc': 'the NWBFile object to export. If None, then the entire contents of src_io will be exported',
             'default': None},
            {'name': 'write_args', 'type': dict,
             'doc': 'arguments to pass to :py:meth:`~hdmf.backends.io.HDMFIO.write_builder`',
             'default': None},
            {'name': 'cache_spec', 'type': bool, 'doc': 'whether to cache the specification to file',
             'default': True})
    def export(self, **kwargs):
        """
        Export an NWB file to a new NWB file using the HDF5 backend.

        If ``nwbfile`` is provided, then the build manager of ``src_io`` is used to build the container,
        and the resulting builder will be exported to the new backend. So if ``nwbfile`` is provided,
        ``src_io`` must have a non-None manager property. If ``nwbfile`` is None, then the contents of
        ``src_io`` will be read and exported to the new backend.

        Arguments can be passed in for the ``write_builder`` method using ``write_args``. Some arguments may not be
        supported during export. ``{'link_data': False}`` can be used to copy any datasets linked to from
        the original file instead of creating a new link to those datasets in the exported file.

        The exported file will not contain any links to the original file. All links, internal and external,
        will be preserved in the exported file. All references will also be preserved in the exported file.

        The exported file will use the latest schema version supported by the version of PyNWB used. For example, if
        the input file uses the NWB schema version 2.1 and the latest schema version supported by PyNWB is 2.3,
        then the exported file will use the 2.3 NWB schema.

        Example usage:

        .. code-block:: python

           with NWBHDF5IO(self.read_path, mode='r') as read_io:
               nwbfile = read_io.read()
               # ...  # modify nwbfile
               nwbfile.set_modified()  # this may be necessary if the modifications are changes to attributes

               with NWBHDF5IO(self.export_path, mode='w') as export_io:
                   export_io.export(src_io=read_io, nwbfile=nwbfile)

        See :ref:`export` and :ref:`modifying_data` for more information and examples.
        """
        nwbfile = popargs('nwbfile', kwargs)
        kwargs['container'] = nwbfile
        super().export(**kwargs)


from . import io as __io  # noqa: F401,E402
from .core import NWBContainer, NWBData  # noqa: F401,E402
from .base import TimeSeries, ProcessingModule  # noqa: F401,E402
from .file import NWBFile  # noqa: F401,E402
from . import behavior  # noqa: F401,E402
from . import device  # noqa: F401,E402
from . import ecephys  # noqa: F401,E402
from . import epoch  # noqa: F401,E402
from . import icephys  # noqa: F401,E402
from . import image  # noqa: F401,E402
from . import misc  # noqa: F401,E402
from . import ogen  # noqa: F401,E402
from . import ophys  # noqa: F401,E402
from . import legacy  # noqa: F401,E402
from hdmf.data_utils import DataChunkIterator  # noqa: F401,E402
from hdmf.backends.hdf5 import H5DataIO  # noqa: F401,E402

try:
    # see https://effigies.gitlab.io/posts/python-packaging-2023/
    from ._version import __version__
except ImportError:  # pragma: no cover
    # this is a relatively slower method for getting the version string
    from importlib.metadata import version  # noqa: E402

    __version__ = version("pynwb")
    del version

from ._due import due, BibTeX  # noqa: E402
due.cite(
    BibTeX("""
@article {10.7554/eLife.78362,
article_type = {journal},
title = {{The Neurodata Without Borders ecosystem for neurophysiological data science}},
author = {R\"ubel, Oliver and Tritt, Andrew and Ly, Ryan and Dichter, Benjamin K and
          Ghosh, Satrajit and Niu, Lawrence and Baker, Pamela and Soltesz, Ivan and Ng,
          Lydia and Svoboda, Karel and Frank, Loren and Bouchard, Kristofer E},
editor = {Colgin, Laura L and Jadhav, Shantanu P},
volume = {11},
year = {2022},
month = {oct},
pub_date = {2022-10-04},
pages = {e78362},
citation = {eLife 2022;11:e78362},
doi = {10.7554/eLife.78362},
url = {https://doi.org/10.7554/eLife.78362},
keywords = {Neurophysiology, data ecosystem, data language, data standard, FAIR data, archive},
journal = {eLife},
issn = {2050-084X},
publisher = {eLife Sciences Publications, Ltd}}
"""),
    description="The Neurodata Without Borders ecosystem for neurophysiological data science",
    path="pynwb/", version=__version__,
    cite_module=True
)
del due, BibTeX



import subprocess
import re
import sys
from unittest.mock import patch
from io import StringIO
import warnings

from pynwb.testing import TestCase
from pynwb import validate, NWBHDF5IO


# NOTE we use "coverage run -m pynwb.validate" instead of "python -m pynwb.validate"
# so that we can both test pynwb.validate and compute code coverage from that test.
# NOTE we also use "coverage run -p" which will generate a .coverage file with the
# machine name, process id, and a random number appended to the filename to
# simplify collecting and merging coverage data from multiple subprocesses. if "-p"
# is not used, then each "coverage run" will overwrite the .coverage file from a
# previous "coverage run".
# NOTE we run "coverage" as "{sys.executable} -m coverage" to 1. make sure to use
# the same python version, and on Debian systems executable is "python3-coverage", not
# just "coverage".
# NOTE the run_coverage.yml GitHub Action runs "python -m coverage combine" to
# combine the individual coverage reports into one .coverage file.
def run_coverage(extra_args: list[str]):
    return subprocess.run(
        [sys.executable, "-m", "coverage", "run", "-p", "-m", "pynwb.validate"]
        + extra_args,
        capture_output=True
    )


class TestValidateCLI(TestCase):

    # 1.0.2_nwbfile.nwb has no cached specifications
    # 1.0.3_nwbfile.nwb has cached "core" specification
    # 1.1.2_nwbfile.nwb has cached "core" and "hdmf-common" specifications

    def test_validate_file_no_cache(self):
        """Test that validating a file with no cached spec against the core namespace succeeds."""
        result = run_coverage(["tests/back_compat/1.0.2_nwbfile.nwb"])

        stderr_regex = re.compile(
            r"The file tests/back_compat/1\.0\.2_nwbfile\.nwb has no cached namespace information\. "
            r"Falling back to PyNWB namespace information\.\s*"
        )
        self.assertRegex(result.stderr.decode('utf-8'), stderr_regex)

        stdout_regex = re.compile(
            r"Validating tests/back_compat/1\.0\.2_nwbfile\.nwb against PyNWB namespace information using namespace "
            r"'core'\.\s* - no errors found\.\s*")
        self.assertRegex(result.stdout.decode('utf-8'), stdout_regex)

    def test_validate_file_no_cache_bad_ns(self):
        """Test that validating a file with no cached spec against a specified, unknown namespace fails."""
        result = run_coverage(["tests/back_compat/1.0.2_nwbfile.nwb", "--ns", "notfound"])

        stderr_regex = re.compile(
            r"The file tests/back_compat/1\.0\.2_nwbfile\.nwb has no cached namespace information\. "
            r"Falling back to PyNWB namespace information\.\s*"
            r"The namespace 'notfound' could not be found in PyNWB namespace information as only "
            r"\['core'\] is present\.\s*"
        )
        self.assertRegex(result.stderr.decode('utf-8'), stderr_regex)

        self.assertEqual(result.stdout.decode('utf-8'), '')

    def test_validate_file_cached(self):
        """Test that validating a file with cached spec against its cached namespace succeeds."""
        result = run_coverage(["tests/back_compat/1.1.2_nwbfile.nwb"])

        self.assertEqual(result.stderr.decode('utf-8'), '')

        stdout_regex = re.compile(
            r"Validating tests/back_compat/1\.1\.2_nwbfile\.nwb against cached namespace information using namespace "
            r"'core'\.\s* - no errors found\.\s*")
        self.assertRegex(result.stdout.decode('utf-8'), stdout_regex)

    def test_validate_file_cached_bad_ns(self):
        """Test that validating a file with cached spec against a specified, unknown namespace fails."""
        result = run_coverage(["tests/back_compat/1.1.2_nwbfile.nwb", "--ns", "notfound"])

        stderr_regex = re.compile(
            r"The namespace 'notfound' could not be found in cached namespace information as only "
            r"\['core'\] is present\.\s*"
        )
        self.assertRegex(result.stderr.decode('utf-8'), stderr_regex)

        self.assertEqual(result.stdout.decode('utf-8'), '')

    def test_validate_file_cached_extension(self):
        """Test that validating a file with cached spec against the cached namespaces succeeds."""
        result = run_coverage(["tests/back_compat/2.1.0_nwbfile_with_extension.nwb"])

        self.assertEqual(result.stderr.decode('utf-8'), '')

        stdout_regex = re.compile(
            r"Validating tests/back_compat/2\.1\.0_nwbfile_with_extension\.nwb against cached namespace information "
            r"using namespace 'ndx-testextension'\.\s* - no errors found\.\s*")
        self.assertRegex(result.stdout.decode('utf-8'), stdout_regex)

    def test_validate_file_cached_extension_pass_ns(self):
        """Test that validating a file with cached spec against the extension namespace succeeds."""
        result = run_coverage(["tests/back_compat/2.1.0_nwbfile_with_extension.nwb", "--ns", "ndx-testextension"])

        self.assertEqual(result.stderr.decode('utf-8'), '')

        stdout_regex = re.compile(
            r"Validating tests/back_compat/2\.1\.0_nwbfile_with_extension\.nwb against cached namespace information "
            r"using namespace 'ndx-testextension'\.\s* - no errors found\.\s*")
        self.assertRegex(result.stdout.decode('utf-8'), stdout_regex)

    def test_validate_file_cached_core(self):
        """Test that validating a file with cached spec against the core namespace succeeds."""
        result = run_coverage(["tests/back_compat/2.1.0_nwbfile_with_extension.nwb", "--ns", "core"])

        stdout_regex = re.compile(
            r"The namespace 'core' is included by the namespace 'ndx-testextension'. "
            r"Please validate against that namespace instead\.\s*"
        )
        self.assertRegex(result.stderr.decode('utf-8'), stdout_regex)

    def test_validate_file_cached_hdmf_common(self):
        """Test that validating a file with cached spec against the hdmf-common namespace fails."""
        result = run_coverage(["tests/back_compat/1.1.2_nwbfile.nwb", "--ns", "hdmf-common"])

        stderr_regex = re.compile(
            r"The namespace 'hdmf-common' is included by the namespace 'core'\. Please validate against that "
            r"namespace instead\.\s*",
        )
        self.assertRegex(result.stderr.decode('utf-8'), stderr_regex)

    def test_validate_file_cached_ignore(self):
        """Test that validating a file with cached spec against the core namespace succeeds."""
        result = run_coverage(["tests/back_compat/1.1.2_nwbfile.nwb", "--no-cached-namespace"])

        self.assertEqual(result.stderr.decode('utf-8'), '')

        stdout_regex = re.compile(
            r"Validating tests/back_compat/1\.1\.2_nwbfile\.nwb against PyNWB namespace information using namespace "
            r"'core'\.\s* - no errors found\.\s*")
        self.assertRegex(result.stdout.decode('utf-8'), stdout_regex)

    def test_validate_file_invalid(self):
        """Test that validating an invalid file outputs errors."""
        result = run_coverage(["tests/back_compat/1.0.2_str_experimenter.nwb", "--no-cached-namespace"])

        stderr_regex = re.compile(
            r" - found the following errors:\s*"
            r"root/general/experimenter \(general/experimenter\): incorrect shape - expected an array of shape "
            r"'\[None\]', got non-array data 'one experimenter'\s*"
        )
        self.assertRegex(result.stderr.decode('utf-8'), stderr_regex)

        stdout_regex = re.compile(
            r"Validating tests/back_compat/1\.0\.2_str_experimenter\.nwb against PyNWB namespace information using "
            r"namespace 'core'\.\s*")
        self.assertRegex(result.stdout.decode('utf-8'), stdout_regex)

    def test_validate_file_list_namespaces_core(self):
        """Test listing namespaces from a file"""
        result = run_coverage(["tests/back_compat/1.1.2_nwbfile.nwb", "--list-namespaces"])

        self.assertEqual(result.stderr.decode('utf-8'), '')

        stdout_regex = re.compile(r"core\s*")
        self.assertRegex(result.stdout.decode('utf-8'), stdout_regex)

    def test_validate_file_list_namespaces_extension(self):
        """Test listing namespaces from a file with an extension"""
        result = run_coverage(["tests/back_compat/2.1.0_nwbfile_with_extension.nwb", "--list-namespaces"])

        self.assertEqual(result.stderr.decode('utf-8'), '')

        stdout_regex = re.compile(r"ndx-testextension\s*")
        self.assertRegex(result.stdout.decode('utf-8'), stdout_regex)


class TestValidateFunction(TestCase):

    # 1.0.2_nwbfile.nwb has no cached specifications
    # 1.0.3_nwbfile.nwb has cached "core" specification
    # 1.1.2_nwbfile.nwb has cached "core" and "hdmf-common" specificaitions

    def get_io(self, path):
        """Get an NWBHDF5IO object for the given path, ignoring the warning about ignoring cached namespaces."""
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=r"Ignoring cached namespace .*",
                category=UserWarning,
            )
            return NWBHDF5IO(str(path), 'r')

    def test_validate_io_no_cache(self):
        """Test that validating a file with no cached spec against the core namespace succeeds."""
        with self.get_io('tests/back_compat/1.0.2_nwbfile.nwb') as io:
            errors = validate(io)
            self.assertEqual(errors, [])

    def test_validate_io_no_cache_bad_ns(self):
        """Test that validating a file with no cached spec against a specified, unknown namespace fails."""
        with self.get_io('tests/back_compat/1.0.2_nwbfile.nwb') as io:
            with self.assertRaisesWith(KeyError, "\"'notfound' not a namespace\""):
                validate(io, 'notfound')

    def test_validate_io_cached(self):
        """Test that validating a file with cached spec against its cached namespace succeeds."""
        with self.get_io('tests/back_compat/1.1.2_nwbfile.nwb') as io:
            errors = validate(io)
            self.assertEqual(errors, [])

    def test_validate_io_cached_extension(self):
        """Test that validating a file with cached spec against its cached namespaces succeeds."""
        with self.get_io('tests/back_compat/2.1.0_nwbfile_with_extension.nwb') as io:
            errors = validate(io)
            self.assertEqual(errors, [])

    def test_validate_io_cached_extension_pass_ns(self):
        """Test that validating a file with cached extension spec against the extension namespace succeeds."""
        with self.get_io('tests/back_compat/2.1.0_nwbfile_with_extension.nwb') as io:
            errors = validate(io, 'ndx-testextension')
            self.assertEqual(errors, [])

    def test_validate_io_cached_core_with_io(self):
        """
        For back-compatability, test that validating a file with cached extension spec against the core
        namespace succeeds when using the `io` + `namespace` keywords.
        """
        with self.get_io(path='tests/back_compat/2.1.0_nwbfile_with_extension.nwb') as io:
            results = validate(io=io, namespace="core")
            self.assertEqual(results, [])

    def test_validate_file_cached_extension(self):
        """
        Test that validating a file with cached extension spec against the core
        namespace raises an error with the new CLI-mimicing paths keyword.
        """
        nwbfile_path = "tests/back_compat/2.1.0_nwbfile_with_extension.nwb"
        with patch("sys.stderr", new=StringIO()) as fake_err:
            with patch("sys.stdout", new=StringIO()) as fake_out:
                results, status = validate(paths=[nwbfile_path], namespace="core", verbose=True)
                self.assertEqual(results, [])
                self.assertEqual(status, 1)
                self.assertEqual(
                    fake_err.getvalue(),
                    (
                        "The namespace 'core' is included by the namespace 'ndx-testextension'. "
                        "Please validate against that namespace instead.\n"
                    )
                )
                self.assertEqual(fake_out.getvalue(), "")

    def test_validate_file_cached_core(self):
        """
        Test that validating a file with cached core spec with verbose=False.
        """
        nwbfile_path = "tests/back_compat/1.1.2_nwbfile.nwb"
        with patch("sys.stderr", new=StringIO()) as fake_err:
            with patch("sys.stdout", new=StringIO()) as fake_out:
                results, status = validate(paths=[nwbfile_path], namespace="core")
                self.assertEqual(results, [])
                self.assertEqual(status, 0)
                self.assertEqual(fake_err.getvalue(), "")
                self.assertEqual(fake_out.getvalue(), "")

    def test_validate_file_cached_no_cache_bad_ns(self):
        """
        Test that validating a file with no cached namespace, a namespace that is not found, and verbose=False.
        """
        nwbfile_path = "tests/back_compat/1.0.2_nwbfile.nwb"
        with patch("sys.stderr", new=StringIO()) as fake_err:
            with patch("sys.stdout", new=StringIO()) as fake_out:
                results, status = validate(paths=[nwbfile_path], namespace="notfound")
                self.assertEqual(results, [])
                self.assertEqual(status, 1)
                stderr_regex = (
                    r"The namespace 'notfound' could not be found in PyNWB namespace information as only "
                    r"\['core'\] is present.\n"
                )
                self.assertRegex(fake_err.getvalue(), stderr_regex)
                self.assertEqual(fake_out.getvalue(), "")

    def test_validate_io_cached_bad_ns(self):
        """Test that validating a file with cached spec against a specified, unknown namespace fails."""
        with self.get_io('tests/back_compat/1.1.2_nwbfile.nwb') as io:
            with self.assertRaisesWith(KeyError, "\"'notfound' not a namespace\""):
                validate(io, 'notfound')

    def test_validate_io_cached_hdmf_common(self):
        """Test that validating a file with cached spec against the hdmf-common namespace fails."""
        with self.get_io('tests/back_compat/1.1.2_nwbfile.nwb') as io:
            # TODO this error should not be different from the error when using the validate script above
            msg = "builder must have data type defined with attribute 'data_type'"
            with self.assertRaisesWith(ValueError, msg):
                validate(io, 'hdmf-common')



import numpy as np
from pathlib import Path
import warnings

from pynwb import NWBHDF5IO, validate, TimeSeries
from pynwb.image import ImageSeries
from pynwb.testing import TestCase


class TestReadOldVersions(TestCase):

    expected_warnings = {
        ''
        '2.1.0_imageseries_non_external_format.nwb': [(
            "ImageSeries 'test_imageseries': Format must be 'external' when external_file is specified."
        )],
        '2.1.0_imageseries_nonmatch_starting_frame.nwb': [(
            "ImageSeries 'test_imageseries': The number of frame indices in 'starting_frame' should have the same "
            "length as 'external_file'."
        )],
    }

    expected_errors = {
        '1.0.2_str_experimenter.nwb': [("root/general/experimenter (general/experimenter): incorrect shape - expected "
                                        "an array of shape '[None]', got non-array data 'one experimenter'")],
        '1.0.3_str_experimenter.nwb': [("root/general/experimenter (general/experimenter): incorrect shape - expected "
                                        "an array of shape '[None]', got non-array data 'one experimenter'")],
        '1.0.2_str_pub.nwb': [("root/general/related_publications (general/related_publications): incorrect shape "
                               "- expected an array of shape '[None]', got non-array data 'one publication'")],
        '1.0.3_str_pub.nwb': [("root/general/related_publications (general/related_publications): incorrect shape "
                               "- expected an array of shape '[None]', got non-array data 'one publication'")],
        '1.5.1_timeseries_no_data.nwb': [("TimeSeries/data/data (acquisition/test_timeseries/data): argument missing")],
        '1.5.1_timeseries_no_unit.nwb': [("TimeSeries/data/unit (acquisition/test_timeseries/data): argument missing")],
        '1.5.1_imageseries_no_data.nwb': [("ImageSeries/data/data (acquisition/test_imageseries/data): "
                                           "argument missing")],
        '1.5.1_imageseries_no_unit.nwb': [("ImageSeries/data/unit (acquisition/test_imageseries/data): "
                                           "argument missing")],
    }

    def get_io(self, path):
        """Get an NWBHDF5IO object for the given path."""
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=r"Ignoring cached namespace .*",
                category=UserWarning,
            )
            return NWBHDF5IO(str(path), 'r')

    def test_read(self):
        """Test reading and validating all NWB files in the same folder as this file.

        This folder contains NWB files generated by previous versions of NWB using the script
        src/pynwb/testing/make_test_files.py
        """
        dir_path = Path(__file__).parent
        nwb_files = dir_path.glob('*.nwb')
        for f in nwb_files:
            with self.subTest(file=f.name):
                with warnings.catch_warnings(record=True) as warnings_on_read:
                    warnings.simplefilter("always")
                    with self.get_io(f) as io:
                        errors = validate(io)
                        io.read()
                        for w in warnings_on_read:
                            if f.name in self.expected_warnings:
                                if str(w.message) not in self.expected_warnings[f.name]:
                                    pass
                                    # will replace above with below after the test file is updated
                                    # raise Exception("Unexpected warning: %s: %s" % (f.name, str(w.message)))
                            else:
                                pass
                                # will replace above with below after the test file is updated
                                # raise Exception("Unexpected warning: %s: %s" % (f.name, str(w.message)))
                        if errors:
                            for e in errors:
                                if f.name in self.expected_errors:
                                    if str(e) not in self.expected_errors[f.name]:
                                        warnings.warn('%s: %s' % (f.name, e))
                                else:
                                    raise Exception("Unexpected validation error: %s: %s" % (f.name, e))
                            # TODO uncomment below when validation errors have been fixed
                            # raise Exception('%d validation error(s). See warnings.' % len(errors))

    def test_read_timeseries_no_data(self):
        """Test that a TimeSeries written without data is read with data set to the default value."""
        f = Path(__file__).parent / '1.5.1_timeseries_no_data.nwb'
        with self.get_io(f) as io:
            read_nwbfile = io.read()
            np.testing.assert_array_equal(read_nwbfile.acquisition['test_timeseries'].data, TimeSeries.DEFAULT_DATA)

    def test_read_timeseries_no_unit(self):
        """Test that an ImageSeries written without unit is read with unit set to the default value."""
        f = Path(__file__).parent / '1.5.1_timeseries_no_unit.nwb'
        with self.get_io(f) as io:
            read_nwbfile = io.read()
            self.assertEqual(read_nwbfile.acquisition['test_timeseries'].unit, TimeSeries.DEFAULT_UNIT)

    def test_read_imageseries_no_data(self):
        """Test that an ImageSeries written without data is read with data set to the default value."""
        f = Path(__file__).parent / '1.5.1_imageseries_no_data.nwb'
        with self.get_io(f) as io:
            read_nwbfile = io.read()
            np.testing.assert_array_equal(read_nwbfile.acquisition['test_imageseries'].data, ImageSeries.DEFAULT_DATA)

    def test_read_imageseries_no_unit(self):
        """Test that an ImageSeries written without unit is read with unit set to the default value."""
        f = Path(__file__).parent / '1.5.1_imageseries_no_unit.nwb'
        with self.get_io(f) as io:
            read_nwbfile = io.read()
            self.assertEqual(read_nwbfile.acquisition['test_imageseries'].unit, ImageSeries.DEFAULT_UNIT)

    def test_read_imageseries_non_external_format(self):
        """Test that reading an ImageSeries with an inconsistent format does not change the value."""
        fbase = "2.1.0_imageseries_non_external_format.nwb"
        f = Path(__file__).parent / fbase
        expected_warning = self.expected_warnings[fbase][0]
        with self.assertWarnsWith(UserWarning, expected_warning):
            with self.get_io(f) as io:
                read_nwbfile = io.read()
                self.assertEqual(read_nwbfile.acquisition['test_imageseries'].format, "tiff")

    def test_read_imageseries_nonmatch_starting_frame(self):
        """Test that reading an ImageSeries with an inconsistent starting_frame does not change the value."""
        fbase = "2.1.0_imageseries_nonmatch_starting_frame.nwb"
        f = Path(__file__).parent / fbase
        expected_warning = self.expected_warnings[fbase][0]
        with self.assertWarnsWith(UserWarning, expected_warning):
            with self.get_io(f) as io:
                read_nwbfile = io.read()
                np.testing.assert_array_equal(read_nwbfile.acquisition['test_imageseries'].starting_frame, [1, 2, 3])

    def test_read_subject_no_age__reference(self):
        """Test that reading a Subject without an age__reference set with NWB schema 2.5.0 sets the value to None"""
        f = Path(__file__).parent / '2.2.0_subject_no_age__reference.nwb'
        with self.get_io(f) as io:
            read_nwbfile = io.read()
            self.assertIsNone(read_nwbfile.subject.age__reference)



from typing import Optional, Sequence

import numpy as np

from hdmf.common.table import DynamicTableRegion

from ... import NWBFile, ProcessingModule
from ...device import Device
from ...ophys import (
    RoiResponseSeries,
    OpticalChannel,
    ImagingPlane,
    OnePhotonSeries,
    TwoPhotonSeries,
    PlaneSegmentation,
    ImageSegmentation,
    DfOverF,
    Fluorescence,
)
from .device import mock_Device
from .utils import name_generator


def mock_OpticalChannel(
    name: Optional[str] = None,
    description: str = "description",
    emission_lambda: float = 500.0,
    nwbfile: Optional[NWBFile] = None,
) -> OpticalChannel:
    optical_channel = OpticalChannel(
        name=name or name_generator("OpticalChannel"), description=description, emission_lambda=emission_lambda,
    )

    if nwbfile is not None:
        mock_ImagingPlane(nwbfile=nwbfile, optical_channel=optical_channel)

    return optical_channel


def mock_ImagingPlane(
    name: Optional[str] = None,
    optical_channel: Optional[OpticalChannel] = None,
    description: str = "description",
    device: Optional[Device] = None,
    excitation_lambda: float = 500.0,
    indicator: str = "indicator",
    location: str = "unknown",
    imaging_rate: float = 30.0,
    manifold=None,
    conversion: float = 1.0,
    unit: str = "meters",
    reference_frame=None,
    origin_coords=None,
    origin_coords_unit: str = "meters",
    grid_spacing=None,
    grid_spacing_unit: str = "meters",
    nwbfile: Optional[NWBFile] = None,
) -> ImagingPlane:
    imaging_plane = ImagingPlane(
        name=name or name_generator("ImagingPlane"),
        optical_channel=optical_channel or mock_OpticalChannel(nwbfile=nwbfile),
        description=description,
        device=device or mock_Device(nwbfile=nwbfile),
        excitation_lambda=excitation_lambda,
        indicator=indicator,
        location=location,
        imaging_rate=imaging_rate,
        manifold=manifold,
        conversion=conversion,
        unit=unit,
        reference_frame=reference_frame,
        origin_coords=origin_coords,
        origin_coords_unit=origin_coords_unit,
        grid_spacing=grid_spacing,
        grid_spacing_unit=grid_spacing_unit,
    )

    if nwbfile is not None:
        if "ophys" not in nwbfile.processing:
            nwbfile.create_processing_module("ophys", "ophys")
        nwbfile.add_imaging_plane(imaging_plane)

    return imaging_plane


def mock_OnePhotonSeries(
    name: Optional[str] = None,
    imaging_plane: Optional[ImagingPlane] = None,
    data=None,
    rate: Optional[float] = 50.0,
    unit: str = "n.a.",
    exposure_time=None,
    binning=None,
    power=None,
    intensity=None,
    format=None,
    pmt_gain=None,
    scan_line_rate=None,
    external_file=None,
    starting_frame=[0],
    bits_per_pixel=None,
    dimension=None,
    resolution=-1.0,
    conversion=1.0,
    offset=0.0,
    timestamps=None,
    starting_time=None,
    comments="no comments",
    description="no description",
    control=None,
    control_description=None,
    device=None,
    nwbfile: Optional[NWBFile] = None,
) -> OnePhotonSeries:
    one_photon_series = OnePhotonSeries(
        name=name if name is not None else name_generator("OnePhotonSeries"),
        imaging_plane=imaging_plane or mock_ImagingPlane(nwbfile=nwbfile),
        data=data if data is not None else np.ones((20, 5, 5)),
        unit=unit,
        exposure_time=exposure_time,
        binning=binning,
        power=power,
        intensity=intensity,
        format=format,
        pmt_gain=pmt_gain,
        scan_line_rate=scan_line_rate,
        external_file=external_file,
        starting_frame=starting_frame,
        bits_per_pixel=bits_per_pixel,
        dimension=dimension,
        resolution=resolution,
        conversion=conversion,
        timestamps=timestamps,
        starting_time=starting_time,
        offset=offset,
        rate=rate,
        comments=comments,
        description=description,
        control=control,
        control_description=control_description,
        device=device,
    )

    if nwbfile is not None:
        nwbfile.add_acquisition(one_photon_series)

    return one_photon_series


def mock_TwoPhotonSeries(
    name: Optional[str] = None,
    imaging_plane: Optional[ImagingPlane] = None,
    data=None,
    rate=50.0,
    unit="n.a.",
    format=None,
    field_of_view=None,
    pmt_gain=None,
    scan_line_rate=None,
    external_file=None,
    starting_frame=[0],
    bits_per_pixel=None,
    dimension=None,
    resolution=-1.0,
    conversion=1.0,
    offset=0.0,
    timestamps=None,
    starting_time=None,
    comments="no comments",
    description="no description",
    control=None,
    control_description=None,
    device=None,
    nwbfile: Optional[NWBFile] = None,
) -> TwoPhotonSeries:
    two_photon_series = TwoPhotonSeries(
        name=name if name is not None else name_generator("TwoPhotonSeries"),
        imaging_plane=imaging_plane or mock_ImagingPlane(nwbfile=nwbfile),
        data=data if data is not None else np.ones((20, 5, 5)),
        unit=unit,
        format=format,
        field_of_view=field_of_view,
        pmt_gain=pmt_gain,
        scan_line_rate=scan_line_rate,
        external_file=external_file,
        starting_frame=starting_frame,
        bits_per_pixel=bits_per_pixel,
        dimension=dimension,
        resolution=resolution,
        conversion=conversion,
        timestamps=timestamps,
        starting_time=starting_time,
        rate=rate,
        comments=comments,
        description=description,
        control=control,
        control_description=control_description,
        device=device,
        offset=offset,
    )

    if nwbfile is not None:
        nwbfile.add_acquisition(two_photon_series)

    return two_photon_series


def mock_PlaneSegmentation(
    description: str = "no description",
    imaging_plane: Optional[ImagingPlane] = None,
    name: Optional[str] = None,
    reference_images=None,
    n_rois: int = 5,
    nwbfile: Optional[NWBFile] = None,
) -> PlaneSegmentation:
    plane_segmentation = PlaneSegmentation(
        description=description,
        imaging_plane=imaging_plane or mock_ImagingPlane(nwbfile=nwbfile),
        name=name if name is not None else name_generator("PlaneSegmentation"),
        reference_images=reference_images,
    )

    for _ in range(n_rois):
        plane_segmentation.add_roi(image_mask=np.zeros((10, 10)))

    if nwbfile is not None:
        if "ophys" not in nwbfile.processing:
            nwbfile.create_processing_module("ophys", "ophys")
        nwbfile.processing["ophys"].add(plane_segmentation)

    return plane_segmentation


def mock_ImageSegmentation(
    plane_segmentations: Optional[Sequence[PlaneSegmentation]] = None,
        name: Optional[str] = None,
        nwbfile: Optional[NWBFile] = None
) -> ImageSegmentation:
    image_segmentation = ImageSegmentation(
        plane_segmentations=plane_segmentations or [mock_PlaneSegmentation(nwbfile=nwbfile)],
        name=name or name_generator("ImageSegmentation"),
    )

    if nwbfile is not None:
        if "ophys" not in nwbfile.processing:
            nwbfile.create_processing_module("ophys", "ophys")

        nwbfile.processing["ophys"].add(image_segmentation)

    return image_segmentation


def mock_RoiResponseSeries(
    name: Optional[str] = None,
    data=None,
    unit: str = "n.a.",
    rois=None,
    resolution=-1.0,
    conversion=1.0,
    timestamps=None,
    starting_time=None,
    rate=50.0,
    comments="no comments",
    description="no description",
    control=None,
    control_description=None,
    n_rois=None,
    plane_segmentation: Optional[PlaneSegmentation] = None,
    nwbfile: Optional[NWBFile] = None,
) -> RoiResponseSeries:
    if data is not None:
        if n_rois is not None and n_rois != data.shape[1]:
            raise ValueError("Argument conflict: n_rois does not match second dimension of data.")
        n_rois = data.shape[1]
    else:
        n_rois = 5

    plane_seg = plane_segmentation or mock_PlaneSegmentation(n_rois=n_rois, nwbfile=nwbfile)

    roi_response_series = RoiResponseSeries(
        name=name if name is not None else name_generator("RoiResponseSeries"),
        data=data if data is not None else np.ones((30, n_rois)),
        unit=unit,
        rois=rois
        or DynamicTableRegion(
            name="rois",
            description="rois",
            table=plane_seg,
            data=list(range(n_rois)),
        ),
        resolution=resolution,
        conversion=conversion,
        timestamps=timestamps,
        starting_time=starting_time,
        rate=rate,
        comments=comments,
        description=description,
        control=control,
        control_description=control_description,
    )

    if nwbfile is not None:
        if "ophys" not in nwbfile.processing:
            nwbfile.create_processing_module("ophys", "ophys")

        if plane_seg.name not in nwbfile.processing["ophys"].data_interfaces:
            nwbfile.processing["ophys"].add(plane_seg)

        nwbfile.processing["ophys"].add(roi_response_series)

    return roi_response_series


def mock_DfOverF(
    roi_response_series: Optional[RoiResponseSeries] = None,
    name: Optional[str] = None,
    nwbfile: Optional[NWBFile] = None
) -> DfOverF:
    df_over_f = DfOverF(
        name=name if name is not None else name_generator("DfOverF"),
    )
    plane_seg = mock_PlaneSegmentation(nwbfile=nwbfile)

    if nwbfile is not None:
        if "ophys" not in nwbfile.processing:
            nwbfile.create_processing_module("ophys", "ophys")

        nwbfile.processing["ophys"].add(df_over_f)

    else:
        pm = ProcessingModule(name="ophys", description="ophys")
        pm.add(plane_seg)
        pm.add(df_over_f)

    df_over_f.add_roi_response_series(
        roi_response_series or mock_RoiResponseSeries(nwbfile=nwbfile, plane_segmentation=plane_seg)
    )
    return df_over_f


def mock_Fluorescence(
    roi_response_series: Optional[Sequence[RoiResponseSeries]] = None,
    name: Optional[str] = None,
    nwbfile: Optional[NWBFile] = None,
) -> Fluorescence:
    fluorescence = Fluorescence(
        name=name if name is not None else name_generator("Fluorescence"),
    )
    plane_seg = mock_PlaneSegmentation(nwbfile=nwbfile)

    if nwbfile is not None:
        if "ophys" not in nwbfile.processing:
            nwbfile.create_processing_module("ophys", "ophys")

        nwbfile.processing["ophys"].add(fluorescence)
    else:
        pm = ProcessingModule(name="ophys", description="ophys")
        pm.add(plane_seg)
        pm.add(fluorescence)

    fluorescence.add_roi_response_series(
        roi_response_series or mock_RoiResponseSeries(nwbfile=nwbfile, plane_segmentation=plane_seg)
    )

    return fluorescence



from hdmf.utils import docval, popargs, get_docval, popargs_to_dict

from . import register_class, CORE_NAMESPACE
from .base import TimeSeries
from .core import NWBContainer
from .device import Device


@register_class('OptogeneticStimulusSite', CORE_NAMESPACE)
class OptogeneticStimulusSite(NWBContainer):
    """Optogenetic stimulus site."""

    __nwbfields__ = ('device',
                     'description',
                     'excitation_lambda',
                     'location')

    @docval({'name': 'name', 'type': str, 'doc': 'The name of this stimulus site.'},
            {'name': 'device', 'type': Device, 'doc': 'The device that was used.'},
            {'name': 'description', 'type': str, 'doc': 'Description of site.'},
            {'name': 'excitation_lambda', 'type': float, 'doc': 'Excitation wavelength in nm.'},
            {'name': 'location', 'type': str, 'doc': 'Location of stimulation site.'})
    def __init__(self, **kwargs):
        args_to_set = popargs_to_dict(('device', 'description', 'excitation_lambda', 'location'), kwargs)
        super().__init__(**kwargs)

        for key, val in args_to_set.items():
            setattr(self, key, val)


@register_class('OptogeneticSeries', CORE_NAMESPACE)
class OptogeneticSeries(TimeSeries):
    '''
    Optogenetic stimulus. The data field is in unit of watts.
    '''

    __nwbfields__ = ('site',)

    @docval(*get_docval(TimeSeries.__init__, 'name'),  # required
            {'name': 'data', 'type': ('array_data', 'data', TimeSeries),  # required
             'shape': [(None, ), (None, None)],
             'doc': 'The data values over time.'},
            {'name': 'site', 'type': OptogeneticStimulusSite,  # required
             'doc': 'The site to which this stimulus was applied.'},
            *get_docval(TimeSeries.__init__, 'resolution', 'conversion', 'timestamps', 'starting_time', 'rate',
                        'comments', 'description', 'control', 'control_description', 'offset'))
    def __init__(self, **kwargs):
        site = popargs('site', kwargs)
        kwargs['unit'] = 'watts'
        super().__init__(**kwargs)
        self.site = site
